/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball.debug;

import java.awt.*;

import hoverball.math.*;


/**
 * Zeichnet einen Kreis oder einen Kreisbogen.
 */
public class Circle extends Debug
{
   private static final int RESOLUTION = hoverball.layout.Layout.RESOLUTION;

   private final Matrix M;
   private final Vector r;
   private final double alpha;
   private final double beta;

   /**
    * Erzeugt ein Kreis-Debug.
    *
    * @param p Mittelpunkt
    * @param r Radius (als sphärischer Winkel)
    */
   public Circle (Vector p, double r) { this(p,r,0,Math.PI*2); }

   /**
    * Erzeugt ein Kreisbogen-Debug mit Startwinkel und Bogenlänge.
    *
    * @param p Mittelpunkt
    * @param r Radius (als sphärischer Winkel)
    * @param alpha Startwinkel
    * @param phi Bogenlänge
    */
   public Circle (Vector p, double r, double alpha, double phi)
   {
      Vector a = vector(p);

      this.M = Sphere.euler(new Vector(new Complex(a).arg(),Math.acos(a.z),alpha));
      this.r = Matrix.rotY(r).c;
      this.alpha = alpha;
      this.beta = alpha + phi;
   }

   /**
    * [Implementierung]
    */
   public void paint (Graphics g, Color color, Color globe, double scale, Sphere sphere, Matrix base, boolean front)
   {
      double z = beta-alpha;
      if (z > 2*Math.PI) z = 2*Math.PI;
      if (z <-2*Math.PI) z =-2*Math.PI;
      int n = 1 + (int)(4 * RESOLUTION*Math.abs(z)/2/Math.PI);  // [!] "4*"?
      z /= n;

      Matrix Z = Matrix.rotZ(z);
      Matrix base_M_Z_M_base = Matrix.conj(Z,M).conj(base);

      Vector x = Vector.mul(r,M).mul(base);
      Point P = null;
      Point X = scale(scale,x);

      for (int i=0; i<n; ++i) {
         x = x.mul(base_M_Z_M_base); /*immutable*/
         P = X;
         X = scale(scale,x);

         if (x.z > 0 ^ front) continue;

         g.setColor(color(color,globe,x.z));
         g.drawLine(P.x,P.y,X.x,X.y);
      }
   }
}
