/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball.debug;

import java.awt.*;

import hoverball.math.*;


/**
 * Zeichnet eine Linie.
 * Die Linie ist die kürzeste Vebindung von zwei Punkten auf der Sphäre.
 */
public class Line extends Debug
{
   private static final int RESOLUTION = hoverball.layout.Layout.RESOLUTION;

   private final int n;
   private final Vector p;
   private final Matrix M;

   /**
    * Erzeugt ein Linien-Debug.
    *
    * @param p Startpunkt
    * @param q Endpunkt
    */
   public Line (Vector p, Vector q)
   {
      Vector a = vector(p);
      Vector b = vector(q);
      double alpha = Math.acos(Vector.dot(a,b));
      Vector axis = Vector.vec(a,b);

      this.n = 1 + (int)(RESOLUTION * alpha/Math.PI/2);
      this.p = a;
      this.M = Matrix.rot(axis,alpha/n);
   }

   /**
    * [Implementierung]
    */
   public void paint (Graphics g, Color color, Color globe, double scale, Sphere sphere, Matrix base, boolean front)
   {
      Matrix base_M_base = Matrix.conj(M,base);

      Vector x = new Vector(p).mul(base);
      Point P = null;
      Point X = scale(scale,x);

      for (int i=0; i<n; ++i) {
         x = x.mul(base_M_base); /*immutable*/
         P = X;
         X = scale(scale,x);

         if (x.z > 0 ^ front) continue;

         g.setColor(color(color,globe,x.z));
         g.drawLine(P.x,P.y,X.x,X.y);
      }
   }
}
