/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball.debug;

import java.awt.*;

import hoverball.math.*;


/**
 * Markiert einen Punkt mit einem X.
 */
public class X extends Debug
{
   private static final int SIZE = 5;
   private static final int SPACING = 14;  // Text's SPACING

   private final Matrix M;
   private final String text;

   /**
    * Erzeugt ein X-Markierungs-Debug.
    *
    * @param p Punkt
    */
   public X (Vector p)
   {
      this(p,null);
   }

   /**
    * Erzeugt ein X-Markierungs-Debug mit bezeichnendem Text.
    *
    * @param p Punkt
    * @param text Text
    */
   public X (Vector p, String text)
   {
      Vector a = vector(p);

      this.M = Sphere.euler(new Vector(new Complex(a).arg(),Math.acos(a.z),0));
      this.text = (text == null)? null : ""+text;
   }

   /**
    * [Implementierung]
    */
   public void paint (Graphics g, Color color, Color globe, double scale, Sphere sphere, Matrix base, boolean front)
   {
      Matrix M_base = Matrix.mul(M,base);
      if (front ^ M_base.c.z > 0) return;

      double x = SIZE/scale;

      Point A = scale(scale,new Vector(+x,+x,1).mul(M_base));
      Point B = scale(scale,new Vector(-x,+x,1).mul(M_base));
      Point C = scale(scale,new Vector(-x,-x,1).mul(M_base));
      Point D = scale(scale,new Vector(+x,-x,1).mul(M_base));

      g.setColor(color(color,globe,M.c.z));
      g.drawLine(A.x,A.y,C.x,C.y);
      g.drawLine(B.x,B.y,D.x,D.y);

      if (text == null) return;

      Point X = scale(scale,M_base.c);
      int width = g.getFontMetrics().stringWidth(text);
      int height = g.getFontMetrics().getAscent();

      g.drawString(text,X.x-width/2,X.y+height/2-SPACING-1);
   }
}

