/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball.net;

import java.io.IOException;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.util.Vector;

public abstract class Server implements Runnable
{
	private ServerSocket serverSocket = null; // TCP network
	private Address serverAddress = null;
	private static final int TIMEOUT = 4000;
	private boolean active = true;

	public abstract void connection (Address sender, String String, Socket socket);
	public abstract void refresh ();
	public boolean multiple; // #hash sensitive?

	public void catchIOException (IOException exception, String where)
	{
	   if (Socket.log!=null) Socket.log.println("Server exception ("+exception+") in "+where);
	}

	public Server (int port) throws IOException
   {
//Socket.log = System.out;               // Network Log (null: no logging)
if (Socket.log != null) Socket.log.println("new Server("+port+")");

	   if (Socket.network==Socket.FULL || Socket.network==Socket.SMART)
	   {
		   try
		   {
		      try {
		          serverSocket = new ServerSocket(port);
		      } 
		      catch (IOException e) {
		    	  serverSocket = new ServerSocket(0); // suche anderen freien Port
		      }
		      if (serverSocket!=null) {
		    	  // [NICHT:] serverAddress = new Address(serverSocket.getInetAddress(),serverSocket.getLocalPort());
		    	  // [EHER:]  serverAddress = new Address(InetAddress.getLocalHost(),serverSocket.getLocalPort());
		    	  serverAddress = new Address(Socket.getMyLocalIP(),serverSocket.getLocalPort());
		      }
		   }
		   catch (IOException e) { catchIOException(e,"constructor"); }
	    }
	    else // Socket.network==Socket.VIRTUAL
	    {
	    	serverAddress = new Address(InetAddress.getLocalHost(),port);
	    }

if (Socket.log != null) Socket.log.println("Server Socket opened: "+serverAddress);
//APPLET THREAD BUG//
	    //new Thread(this).start(); //new Thread(Socket.group,this,"SERVER").start();
	  	// Thread XY = new Thread(this,"Hoverball Server Socket at "+serverAddress);
	  	// XY.start(); // das muss so umständlich sein. (!?)
	  	new Thread(this,"Hoverball Server Socket at "+serverAddress).start();
   }

   public synchronized void close ()
   {
if (Socket.log != null) Socket.log.println("Server.close()");
      active = false;
      try {
    	  if (serverSocket!=null) serverSocket.close();
      } catch (IOException e) { catchIOException(e,"close");}
   }

   public Address getAddress ()
   {
	   return new Address(serverAddress);
   }


	public void run()
	{
		if (Socket.network==Socket.FULL || (Socket.network==Socket.SMART))
		{
if (Socket.log != null) Socket.log.println("socket.run() - start");
		   try
		   {
				while (active)
				{
if (Socket.log != null) Socket.log.println("socket.run() - while active ... accept()");
					java.net.Socket connection = serverSocket.accept();
if (Socket.log != null) Socket.log.println("socket.run() - accepted!");
					new Thread("Hoverball Server -create- "+connection) {
						public void run() {
							try {
if (Socket.log != null) Socket.log.println("socket.run() - run() - start");
								Socket socket = new Socket(connection);
if (Socket.log != null) Socket.log.println("socket.run() - run() - new Socket ... receive()");
								socket.setTimeout(TIMEOUT);
								Message connectData = socket.receive();  // receive first message on connecting
								socket.setTimeout(0);
if (Socket.log != null) Socket.log.println("socket.run() - run() - new Socket ... received: "+connectData);
								connection(connectData.sender,connectData.data,socket);
if (Socket.log != null) Socket.log.println("socket.run() - run() - new Socket ... connection establishes.");

                                                                         // and now: run! (you are already in a new client thread... :-)
                             } catch (IOException e) { catchIOException(e,"run_run"); }
						}
					}.start();
				}
			}
			catch (IOException e) { catchIOException(e,"run"); }
if (Socket.log != null) Socket.log.println("socket.run() leave! ----------------");
		}
		else // (Socket.network==Socket.VIRTUAL)
		{
		      Message message = null;
		      Vector vector = Socket.getLocal(new Connection(serverAddress,serverAddress));
		      if (vector != null)
		      while (active)
		      {
	    		 while (true) {
	    			 if (!vector.isEmpty()) break;  // read packet!

	    			 synchronized(vector) {
			            try {
			                vector.wait(0);
			            } catch (InterruptedException e) {}
			         }
		    	  }
		 		  message = (Message)vector.elementAt(0);
	    		  vector.removeElementAt(0);
	    		  
			      Socket socket = new Socket(serverAddress,new Address(message.data));
			      try {
			    	  socket.setTimeout(TIMEOUT);
			    	  Message connectData = socket.receive();  // waiting for reply ...
			    	  socket.setTimeout(0);
			    	  connection(connectData.sender,connectData.data,socket);
			      } catch (IOException e) { catchIOException(e,"run"); }
		     }
		 }
	}
}
