/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball.serial;

import hoverball.math.Vector;


// ---- Absolute Disk Info
// (ball t n x1 x2 x3 v1 v2 v3 r m Q)
// (unit t n x1 x2 x3 v1 v2 v3 r m Q L R E py "")

public class DiskAbsolute extends DiskInfo
{
   public int what;       // BALL,UNIT
   public int t;          // team number
   public int n;          // disk number in team

   public Vector x;       // position (euler)
   public Vector v;       // impulsion (euler)
   public double r;       // radius
   public double m;       // mass
   public double Q;       // charge
   public double L;       // engine left
   public double R;       // engine right
   public double E;       // energy
   public double py;      // penalty time

   public String message; // message


   public Object clone ()
   {
      return new DiskAbsolute(this);
   }

   public DiskAbsolute (DiskAbsolute disk)
   {
      this(disk.what,disk.t,disk.n,
           disk.x,disk.v,
           disk.r,disk.m,disk.Q,disk.L,disk.R,
           disk.E,disk.py,
           disk.message);
   }

   public DiskAbsolute (int what, int t, int n,
                        Vector x, Vector v,
                        double r, double m, double Q, double L, double R,
                        double E, double py,
                        String message)
   {
      this.what = what;
      this.t = t;
      this.n = n;

      this.x = new Vector(x);
      this.v = new Vector(v);
      this.r = r;
      this.m = m;
      this.Q = Q;
      this.L = L;
      this.R = R;

      this.E = E;
      this.py = py;

      this.message = (message == null)? null: new String(message);
   }

   public void serialize (StringBuffer buffer)
   {
      buffer.append('(');
      buffer.append(putWhat(what));
      buffer.append(' ');
      buffer.append(putInt(t));
      buffer.append(' ');
      buffer.append(putInt(n));
      buffer.append(' ');
      buffer.append(putDouble(x.x));
      buffer.append(' ');
      buffer.append(putDouble(x.y));
      buffer.append(' ');
      buffer.append(putDouble(x.z));
      buffer.append(' ');
      buffer.append(putDouble(v.x));
      buffer.append(' ');
      buffer.append(putDouble(v.y));
      buffer.append(' ');
      buffer.append(putDouble(v.z));
      buffer.append(' ');
      buffer.append(putDouble(r));
      buffer.append(' ');
      buffer.append(putDouble(m));
      buffer.append(' ');
      buffer.append(putDouble(Q));
      if (what == UNIT) {
         buffer.append(' ');
         buffer.append(putDouble(L));
         buffer.append(' ');
         buffer.append(putDouble(R));
         buffer.append(' ');
         buffer.append(putDouble(E));
         buffer.append(' ');
         buffer.append(putDouble(py));
         buffer.append(' ');
         buffer.append(putString(message));
      }
      buffer.append(')');
   }

   public static Serial deserialize (Object[] objects)
   {
      String keyword = getString(objects[0]);
      int what = getWhat(keyword); if (what == UNKNOWN) return null;

      int t = getInt(objects[1]);
      int n = getInt(objects[2]);
      double x1 = getDouble(objects[3]);
      double x2 = getDouble(objects[4]);
      double x3 = getDouble(objects[5]);
      Vector x = new Vector(x1,x2,x3);
      double v1 = getDouble(objects[6]);
      double v2 = getDouble(objects[7]);
      double v3 = getDouble(objects[8]);
      Vector v = new Vector(v1,v2,v3);
      double r = getDouble(objects[9]);
      double m = getDouble(objects[10]);
      double Q = getDouble(objects[11]);
      double L = (what == UNIT)? getDouble(objects[12]) : 0;
      double R = (what == UNIT)? getDouble(objects[13]) : 0;
      double E = (what == UNIT)? getDouble(objects[14]) : 0;
      double py = (what == UNIT)? getDouble(objects[15]) : 0;
      String message = (what == UNIT)? getString(objects[16]) : null;

      return new DiskAbsolute(what,t,n,x,v,r,m,Q,L,R,E,py,message);
   }
}
