/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball.serial;

import java.io.*;
import java.util.*;


public abstract class Protocol
{
   public static String serialize (Serial serial)  // doesn't need to be overridden.
   {
      return serial.serialize();
   }

   public static Serial deserialize (String string)  // to be overridden!
   {
      return null;
   }


   // ---- static Bracket Parser

   public static Object[] parse (String string)
   {
      StringReader reader = new StringReader("("+string.trim()+")");
      try {
         if (reader.read() != '(') return null;
         Object[] objects = parse(reader);
         if (reader.read() == -1) return objects;
      } catch (IOException e) {
      }
      return null;
   }

   private static Object[] parse (StringReader reader)
      throws IOException
   {
      Vector vector = new Vector(16,8);
      String token = "";
      boolean tokenize = false;
      boolean slash = false;
      boolean quote = false;

      while (true) {
         int h = reader.read();

         if (h == -1) throw new IOException("Syntax Error");

         if (!slash) {
            if (h == '\"') { tokenize = true; quote = !quote; continue; }
            if (h == '\\') { tokenize = true; slash = true; continue; }

            if (!quote && (h == ' ' || h == '(' || h == ')')) {
               if (tokenize) { vector.addElement(token); token = ""; tokenize = false; }
               if (h == '(')
                  vector.addElement(parse(reader));
               if (h == ')') {
                  Object[] objects = new Object[vector.size()];
                  vector.copyInto(objects);
                  return objects;
               }
               continue;
            }
         }

         tokenize = true;
         token += (char)h;
         slash = false;
      }
   }
}

