/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball.serial;

import hoverball.net.Address;

public class US extends Protocol // Unit's Client to Server
{
   // Protocol

   public static Serial deserialize (String string)
   {
      Object[] objects = parse(string); if (objects == null) return null;
      Serial serials = new Serial(){};  // Start-Serial (will be discarded.)
      Serial last = serials;

      for (int i=0; i<objects.length; ++i)
      try {
         Object[] subobjects = (Object[])objects[i];
         Serial serial = null;
         if (((serial = US.Connect.deserialize(subobjects)) != null) ||
             ((serial = US.Look.deserialize(subobjects)) != null) ||
             ((serial = US.Action.deserialize(subobjects)) != null) ||
             ((serial = US.Ping.deserialize(subobjects)) != null) ||
             ((serial = US.Bye.deserialize(subobjects)) != null)) last = last.next = serial;
         else throw new Serial.UnknownCommand();
      } catch (Exception e) {
////System.out.println("US.deserialize failed ("+e+"): |"+string+"|");    	  
         return null;
      }
      return serials.next;
   }


   // (connect #hash "The Ultimates!" Robocop_XXL 123456)

   public static class Connect extends Serial
   {
      public String hash;
      public String team;
      public String name;
      public int color;

      public Object clone ()
      {
         return new Connect(clone(hash),clone(team),clone(name),color).next(clone(next));
      }

      public Connect (String hash, String team, String name, int color)
      {
         this.hash = hash;
         this.team = team;
         this.name = name;
         this.color = color;
      }

      public void serialize (StringBuffer buffer)
      {
         buffer.append("(connect ");
         buffer.append(putString(Address.HASH + hash));
         buffer.append(' ');
         buffer.append(putString(team));
         buffer.append(' ');
         buffer.append(putString(name));
         buffer.append(' ');
         buffer.append(putInt(color));
         buffer.append(')');
      }

      public static Serial deserialize (Object[] objects)
      {
         String keyword = getString(objects[0]); if (!keyword.equals("connect")) return null;
         if (objects.length != 5) throw new IllegalNumberOfArguments();

         String hash = getString(objects[1]).substring(1);
         String team = getString(objects[2]);
         String name = getString(objects[3]);
         int color = getInt(objects[4]);
         return new Connect(hash,team,name,color);
      }
   }

   // (look)

   public static class Look extends Serial
   {
      public Object clone () { return new Look().next(clone(next)); }

      public void serialize (StringBuffer buffer) { buffer.append("(look)"); }

      public static Serial deserialize (Object[] objects)
      {
         String keyword = getString(objects[0]); if (!keyword.equals("look")) return null;
         if (objects.length != 1) throw new IllegalNumberOfArguments();

         return new Look();
      }
   }

   // (action * 0.4 -0.2 [hennähh])

   public static class Action extends Serial
   {
      public double polarize;
      public double left;      //  * = Double.NaN
      public double right;
      public String message;   //  weglassen = null

      public Object clone ()
      {
         return new Action(polarize,left,right,message).next(clone(next));
      }

      public Action (double polarize, double left, double right, String message)
      {
         this.polarize = polarize;
         this.left = left;
         this.right = right;
         this.message = (message == null)? null : new String(message);
      }

      public void serialize (StringBuffer buffer)
      {
         buffer.append("(action ");
         buffer.append(putDouble(polarize));
         buffer.append(' ');
         buffer.append(putDouble(left));
         buffer.append(' ');
         buffer.append(putDouble(right));
         if (message != null) {
            buffer.append(' ');
            buffer.append(putString(message));
         }
         buffer.append(')');
      }

      public static Serial deserialize (Object[] objects)
      {
         String keyword = getString(objects[0]); if (!keyword.equals("action")) return null;
         if (objects.length != 4 && objects.length != 5) throw new IllegalNumberOfArguments();

         double polarize = getDouble(objects[1]);
         double left = getDouble(objects[2]);
         double right = getDouble(objects[3]);
         String message = (objects.length == 5)? getString(objects[4]) : null;
         return new Action(polarize,left,right,message);
      }
   }

   // (bye)

   public static class Bye extends Serial
   {
      public Object clone () { return new Bye().next(clone(next)); }

      public void serialize (StringBuffer buffer) { buffer.append("(bye)"); }

      public static Serial deserialize (Object[] objects)
      {
         String keyword = getString(objects[0]); if (!keyword.equals("bye")) return null;
         if (objects.length != 1) throw new IllegalNumberOfArguments();

         return new Bye();
      }
   }

   // (ping)

   public static class Ping extends Serial
   {
      public Object clone () { return new Ping().next(clone(next)); }

      public void serialize (StringBuffer buffer) { buffer.append("(ping)"); }

      public static Serial deserialize (Object[] objects)
      {
         String keyword = getString(objects[0]); if (!keyword.equals("ping")) return null;
         if (objects.length != 1) throw new IllegalNumberOfArguments();

         return new Ping();
      }
   }
}

