/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball.serial;

import java.util.Vector;


// ---- View (=SC.View, =CS.View)
// (view 23.7 [<AbsoluteObject>|<Team>|<Event>]*)

public class View extends Serial
{
   public long arrival = 0;       // [only important for Graphical Debugging systems]

   public double time;            // current time (no time: Double.NaN)
   public DiskAbsolute[] disks;   // information of disks (!)
   public Team[] teams;           // information of scores
//   public Event[] events;         // current events (collisions, flashs)

   public Object clone ()
   {
      DiskAbsolute[] disks2 = new DiskAbsolute[disks.length];
      for (int i=0; i<disks.length; ++i) disks2[i] = (DiskAbsolute)disks[i].clone();
      Team[] teams2 = new Team[teams.length];
      for (int i=0; i<teams.length; ++i) teams2[i] = (Team)teams[i].clone();

      return new View(time,disks2,teams2).next(clone(next));
   }

   public View (double time, DiskAbsolute[] disks, Team[] teams)
   {
      this.time = time;
      this.disks = (disks != null)? disks : new DiskAbsolute[0];
      this.teams = (teams != null)? teams : new Team[0];
//      this.events = (events != null)? events : new Event[0];
   }

   public void serialize (StringBuffer buffer)
   {
      buffer.append("(view");

      if (!Double.isNaN(time)) {
         buffer.append(' ');
         buffer.append(putDouble(time));
      }
      for (int i=0; i<disks.length; ++i) {
         buffer.append(' ');
         disks[i].serialize(buffer);
      }
      for (int i=0; i<teams.length; ++i) {
         buffer.append(' ');
         teams[i].serialize(buffer);
      }
//      for (int i=0; i<events.length; ++i) {
//         buffer.append(' ');
//         events[i].serialize(buffer);
//      }
      buffer.append(')');
   }

   public static Serial deserialize (Object[] objects)
   {
      String keyword = getString(objects[0]); if (!keyword.equals("view")) return null;

      boolean set_time = false;
      double time = Double.NaN;

      if (objects.length > 1 && objects[1] instanceof String) {
         time = getDouble(objects[1]);
         set_time = true;
      }

      int less = 1 + (set_time? 1 : 0);

      Vector vector_disks = new Vector(objects.length-less);
      Vector vector_teams = new Vector(objects.length-less);
//      Vector vector_events = new Vector(objects.length-less);

      for (int i=0; i<objects.length-less; ++i) {
         DiskAbsolute disk = (DiskAbsolute)DiskAbsolute.deserialize((Object[])objects[i+less]);
         if (disk != null) { vector_disks.addElement(disk); continue; }
         Team team = (Team)Team.deserialize((Object[])objects[i+less]);
         if (team != null) { vector_teams.addElement(team); continue; }
//         Event event = (Event)Event.deserialize((Object[])objects[i+less]);
//         if (event != null) { vector_events.addElement(event); continue; }

         throw new WrongObjectListElement();
      }

      DiskAbsolute[] disks = new DiskAbsolute[vector_disks.size()]; vector_disks.copyInto(disks);
      Team[] teams = new Team[vector_teams.size()]; vector_teams.copyInto(teams);
//      Event[] events = new Event[vector_events.size()]; vector_events.copyInto(events);

      return new View(time,disks,teams);
   }
}

