/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball.simulator;

import java.util.Hashtable;


public class Constants
{
   public static final int T_MAX = 99;
   public static final int N_MAX = 99;

   // ---- Index-Finals -------------------------------------------------------

   private static int i = 0;

   public static final int SIMULATOR_FREQUENCY = (i++);
   public static final int SIMULATOR_TIME = (i++);
   public static final int SIMULATOR_PRECISION = (i++);

   public static final int GAME_DURATION = (i++);
   public static final int GAME_SHOTBALLS = (i++);
   public static final int GAME_TEAMBALLS = (i++);
   public static final int GAME_TIMEOUT = (i++);
   public static final int GAME_PENALTY = (i++);
   public static final int GAME_RECHARGE = (i++);

   public static final int WORLD_RADIUS = (i++);
   public static final int WORLD_VISCOSITY = (i++);
   public static final int WORLD_BOUNDARY = (i++);
   public static final int WORLD_PERMITTIVITY = (i++);

   public static final int UNIT_RADIUS = (i++);
   public static final int UNIT_MASS = (i++);
   public static final int UNIT_CHARGE_MIN = (i++);
   public static final int UNIT_CHARGE_MAX = (i++);
   public static final int UNIT_CHARGE_POS = (i++);
   public static final int UNIT_CHARGE_COST = (i++);
   public static final int UNIT_ENGINE_MIN = (i++);
   public static final int UNIT_ENGINE_MAX = (i++);
   public static final int UNIT_ENGINE_POS = (i++);
   public static final int UNIT_ENGINE_COST = (i++);
   public static final int UNIT_ENERGY_MAX = (i++);
   public static final int UNIT_VISION = (i++);
   public static final int UNIT_MESSAGE = (i++);

   public static final int BALL_RADIUS = (i++);
   public static final int BALL_MASS = (i++);
   public static final int BALL_CHARGE = (i++);
   public static final int BALL_HALFLIFE = (i++);

   public static final int CONSTANTS = i;


   // ---- Constant Table -----------------------------------------------------

   public static final char UNIT_SEPARATOR = ':';

   private static final String[][] TABLE = new String[][]
   {
      new String[]{ " ", "simulator.frequency", "50",      "Simulation frequency : Hz" },
      new String[]{ "x", "simulator.time",      "1",       "Real time coefficient : Hz" },
      new String[]{ " ", "simulator.precision", "1",       "Precision coefficient" },

      new String[]{ " ", "game.duration",       "300",     "Game duration" },
      new String[]{ " ", "game.balls.shot",     "1",       "Number of shot balls" },
      new String[]{ " ", "game.balls.team",     "1",       "Number of team balls per team" },
      new String[]{ " ", "game.timeout",        "1",       "Score timeout" },
      new String[]{ " ", "game.penalty",        "10",      "Penalty duration" },
      new String[]{ " ", "game.recharge",       "0.2",     "Energy recharge rate for units" },

      new String[]{ "x", "world.radius",        "50",      "Sphere radius" },
      new String[]{ "x", "world.viscosity",     "0.1",     "Friction viscosity" },
      new String[]{ "x", "world.boundary",      "0.1",     "Friction boundary layer" },
      new String[]{ "x", "world.permittivity",  "0.000001","Electric field permittivity" },

      new String[]{ "x", "unit.radius",         "2",       "Unit radius" },
      new String[]{ "x", "unit.mass",           "4",       "Unit mass" },
      new String[]{ "x", "unit.charge.min",     "-1",      "Unit's highest negative charge" },
      new String[]{ "x", "unit.charge.max",     "1",       "Unit's highest positive charge" },
      new String[]{ "x", "unit.charge.pos",     "0.5",     "Position coefficient of the charging point" },
      new String[]{ "x", "unit.charge.cost",    "10",      "Energy cost for charge" },
      new String[]{ "x", "unit.engine.min",     "-50",     "Unit's highest negative engine force" },
      new String[]{ "x", "unit.engine.max",     "50",      "Unit's highest positive engine force" },
      new String[]{ "x", "unit.engine.pos",     "0.5",     "Position coefficient of the propulsion points" },
      new String[]{ "x", "unit.engine.cost",    "0.001",   "Energy cost for engine force" },
      new String[]{ "x", "unit.energy.max",     "1",       "Unit's highest energy level" },
      new String[]{ "x", "unit.vision",         "1.0",     "Unit's vision angle : * pi" },
      new String[]{ "x", "unit.message",        "100",     "Length of unit messages" },

      new String[]{ "x", "ball.radius",         "1",       "Ball radius" },
      new String[]{ "x", "ball.mass",           "1",       "Ball mass" },
      new String[]{ "x", "ball.charge",         "1",       "A shot ball's permanent charge" },
      new String[]{ "x", "ball.halflife",       "0.005",   "A team ball's charge halflife" }
   };

   public static final int IS_FOR_UNIT;

   static {
      int n=0; for (int k=0; k<CONSTANTS; ++k) if (isForUnit(k)) ++n;
      IS_FOR_UNIT = n;
   }


   // ---- Constant Keys ------------------------------------------------------

   public static boolean isForUnit     (int constant) { return TABLE[constant][0].equals("x"); }
   public static String getKey         (int constant) { return ""+TABLE[constant][1]; }
   public static String getDefault     (int constant) { return ""+TABLE[constant][2]; }
   public static String getExplanation (int constant) { return ""+TABLE[constant][3]; }

   public static int indexOf (String key)
   {
      for (int i=0; i<CONSTANTS; ++i)
         if (getKey(i).equals(key)) return i;
      return -1;
   }


   // ---- Constant Hashtable -------------------------------------------------

   private Hashtable hash = new Hashtable();

   public Constants ()
   {
      setup();  // Setup
   }

   private Constants (Hashtable hash) { this.hash = (Hashtable)hash.clone(); }

   public Object clone ()
   {
      return new Constants(hash);
   }

   public void setup ()
   {
      hash.clear();

      for (int i=0; i<CONSTANTS; ++i) {  // Setup Defaults
         String key = getKey(i);
         String value = getDefault(i);
         hash.put(key,value);
      }
   }


   // ---- Hash Handling ------------------------------------------------------

   public String get (String key)
   {
      return ""+hash.get(key);
   }

   public String put (String key, String value)
   {
      int i = indexOf(key);
      if (i >= 0 && !match(i,value)) value = get(key);
      hash.put(key,value);
      return value;
   }

   public String remove (String key)
   {
      hash.remove(key);
      String value = null;
      int i = indexOf(key);
      if (i >= 0) hash.put(key,value = getDefault(i));
      return value;
   }


   // ---- Value Matching -----------------------------------------------------

   private boolean match (int constant, String value)
   {
      if (value == null) return false;

      // parsed??

      double x = 0;
      try {
         x = Double.valueOf(value).doubleValue();
      } catch(NumberFormatException e) { return false; }

      // bounds???

      if (!((constant == SIMULATOR_FREQUENCY && x >  0) ||
            (constant == SIMULATOR_TIME      && x >= 0) ||
            (constant == SIMULATOR_PRECISION && x > 0) ||

            (constant == GAME_DURATION       && x >= 0) ||
            (constant == GAME_SHOTBALLS      && x >= 0 && x <= N_MAX) ||
            (constant == GAME_TEAMBALLS      && x >= 0 && x <= N_MAX) ||
            (constant == GAME_TIMEOUT        && x > 0) ||
            (constant == GAME_PENALTY        && x >= 0) ||
            (constant == GAME_RECHARGE       && x >= 0) ||

            (constant == WORLD_RADIUS        && x >  0) ||
            (constant == WORLD_VISCOSITY     && x >= 0 && x <= 1) ||
            (constant == WORLD_BOUNDARY      && x >  0) ||
            (constant == WORLD_PERMITTIVITY  && x >  0) ||

            (constant == UNIT_RADIUS         && x >  0) ||
            (constant == UNIT_MASS           && x >  0) ||
            (constant == UNIT_CHARGE_MIN     && x <= 0) ||
            (constant == UNIT_CHARGE_MAX     && x >= 0) ||
            (constant == UNIT_CHARGE_POS     && x >= 0 && x <= 1) ||
            (constant == UNIT_CHARGE_COST    && x >= 0) ||
            (constant == UNIT_ENGINE_MIN     && x <= 0) ||
            (constant == UNIT_ENGINE_MAX     && x >= 0) ||
            (constant == UNIT_ENGINE_POS     && x >= 0 && x <= 1) ||
            (constant == UNIT_ENGINE_COST    && x >= 0) ||
            (constant == UNIT_ENERGY_MAX     && x >= 0) ||
            (constant == UNIT_VISION         && x >= 0 && x <= 2) ||
            (constant == UNIT_MESSAGE        && x >= 0) ||

            (constant == BALL_RADIUS         && x >  0) ||
            (constant == BALL_MASS           && x >  0) ||
            (constant == BALL_CHARGE         && x >= 0) ||
            (constant == BALL_HALFLIFE       && x >  0))) return false;

      return true;
   }

   // ---- Get Constant Value -------------------------------------------------

   public double getConstant (int constant)
   {
      String value = (String)get(getKey(constant));
      try {
         return Double.valueOf(value).doubleValue();
      } catch(NumberFormatException f) { return Double.NaN; }  // should not happen.
   }
}

