/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball.simulator;

import hoverball.net.*;
import hoverball.serial.*;

import java.util.Vector;
import java.util.Enumeration;


// ---- Network Protocol: Control Client to Simulator ---------- [Control] ----


class Control extends NetAgent
{
   private Simulator simulator = null;

   public Control (Socket socket, Simulator simulator)
   {
      super(socket);
      this.simulator = simulator;
   }


   // ---- Receive ---------------------------------------------------------

   public void receive (Address sender, String message)
   {
//	  System.out.println("Control receives: "+message);
      Serial serials = CS.deserialize(message);
      if (serials == null) { send(new SC.Error(message)); return; }

      Serial serial_ = serials;
      while (serial_ != null) {
         Serial serial = serial_;
         serial_ = serial_.next;
         serial.next = null;

         if (serial instanceof CS.Set) simulator.set((CS.Set)serial);
         else if (serial instanceof CS.Channel) simulator.channel((CS.Channel)serial);
         else if (serial instanceof CS.State) simulator.state((CS.State)serial);
         else if (serial instanceof View) simulator.view(this,(View)serial);
         else if (serial instanceof CS.Ping) simulator.ping(this);
         else if (serial instanceof CS.Bye) simulator.bye(this);
      }
   }


// ---- Control Manager -------------------------------- [Control.Manager] ----


   static class Manager  // Control.Manager
   {
      private Simulator simulator = null;
      private Vector vector = new Vector(5,10);

      public Manager (Simulator simulator) { this.simulator = simulator; }


      public synchronized Control create (Address host, Socket socket)
      {
         Control control = new Control(socket,simulator);

         vector.addElement(control);

         return control;
      }

      public synchronized void remove (Control control)
      {
         vector.removeElement(control);
      }

      public synchronized Control[] toArray ()
      {
         Control[] array = new Control[vector.size()];
         vector.copyInto(array);
         return array;
      }

      public synchronized void send (Serial message)
      {
         for (Enumeration<Control> e = vector.elements(); e.hasMoreElements(); ) {
            Control control = (Control)e.nextElement();
            control.send(message);
         }
      }
   }

}
