/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball.simulator;


// ---- Simulator Tool: Timer ------------------------------------ [Timer] ----


class Timer
{
//---- Nicht mehr nötig:
   private final double ratio    = 1;                         // 1/sec
   private final double duration = Double.POSITIVE_INFINITY;  // Game-Time
//----

   private final double frequency;
   private final Runnable task;

   private java.util.Timer timer = null;

   private long time = 0;          // millis


   public Timer (Runnable task, double frequency)
   {
      this.task = task;
      this.frequency = frequency;
   }


   public synchronized double get ()
   {
      if (timer != null)   // then time means: zeitpunkt des starts
         return Math.min(duration,ratio*((double)(System.currentTimeMillis()-time))/1000);
      else                 // then time means: dauer bisherigen spiels
         return Math.min(duration,ratio*((double)time)/1000);
   }

   public synchronized void put (double time)
   {
      if (timer != null) {
         timeswap();
         this.time = (long)(1/ratio*Math.max(0,Math.min(duration,time))*1000);
         timeswap();
      } else
         this.time = (long)(1/ratio*Math.max(0,Math.min(duration,time))*1000);
   }

   private void timeswap () { time = System.currentTimeMillis()-time; }

   private boolean first = false; private boolean first_ = false;  // "xxx_" konstant in einem run()
   private boolean last  = false; private boolean last_  = false;

   public boolean first () { return first_; }
   public boolean last ()  { return last_;  }

   public synchronized void start ()
   {
      if (timer != null) return;
      first = true;

      timer = new java.util.Timer("Hoverball Simulation Timer"); // new java.util.Timer("Simulation Timer");
      timer.schedule(new java.util.TimerTask()
      {
         public void run () {
            first_ = first;
            last_  = last;

            if (first())
               timeswap();

            task.run();

            if (first())
               first_ = first = false;

            if (last()) {
               timeswap();

               timer.cancel();
               timer.purge();
               timer = null;

               last_ = last = false;
            }
         }
      },0,(int)(1000/frequency));
   }

   public synchronized void stop ()
   {
      if (timer == null) return;
      last = true;
   }
}
