package hoverball.team.Barbarians_;

import hoverball.*;
import hoverball.math.*;
import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;
// import Timm.StoredNode;
import hoverball.debug.*;

class BarbarianDefender extends Unit
{
   public BarbarianDefender(String name)
   {
	   super(null,"Aragorn",0xFF4400);
	   state = State.THINK;
   }

   Map<String, StoredNode> storedPos = new HashMap<String, StoredNode>();
   // available states of Aragorn:
   public enum State {THINK, AIM, GET_BALL, RECOVER }

   private State state;
   private double Qmax, Fmax;
   private Sphere sphere;
   private Vector target;
   private double targetDist;
   private String status;
   private double touchingDist;
   private double controlDist;
   private double puckDist;
   private double controlShootDist;
   private double controlAngle;
   private double maxShotDist;
   private double minShotDist;
   private double ballDist;


   public void loop()
   {
	   //Initialization:
	  sphere = new Sphere(option("world.radius"));
      Qmax = option("unit.charge.max");
      Fmax = option("unit.engine.max");
      touchingDist = option("unit.radius") + 2*option("ball.radius");

      controlDist = touchingDist + 4;

      controlShootDist = touchingDist + 1;
      controlAngle = Math.PI/6; // +-30°: 60° control range
      //maxShotDist = option("world.radius")*Math.PI*0.5;
      maxShotDist = 45;
      minShotDist = 20;

      while(look())
      {
          checkOrientation();
          if(storedPos.containsKey("ball 1 1"))
				debug(new Line(self.X.c, storedPos.get("ball 1 1").position.c), -1 );

    	  switch (state)
    	  {
			case GET_BALL:
				get_ball();
				break;
			case AIM:
				precise_kick();
				break;
			case RECOVER:
				recover();
				break;
			default:
				think();
				break;
		 }


    	  targetDist = Math.floor(targetDist * 100.0) / 100.0;
    	  String energyStr = Double.toString(Math.floor(energy*Qmax*100.0) /100.0);
    	  debug(new Text(self.X.c, "Aragorn \n E="+energyStr+"\n status: "+status+"\n d="+targetDist , 1), -1);
    	  if (target != null)
    		  debug(new Line(self.X.c, target));


      } // end of while(look())

   } // end of loop()

   private void think()
   {
	   Puck ball = puck(BALL, 0, 1);
	   Puck goal = puck(BALL, self.t, 1);
	   target = null;
	   targetDist = 0;
	   status = "";

	   if(goal != null && ball != null)
	   {
		   debug(new Circle(goal.X.c, 0.7));//3.0/180*Math.PI));
		   ballDist = sphere.diff(goal.X.c, ball.X.c);
	   }
	   if(storedPos.containsKey("ball 1 1") && ball != null)
	   {
		   debug(new Circle(storedPos.get("ball 1 1").position.c, 0.7));
		   ballDist = sphere.diff(storedPos.get("ball 1 1").position.c, ball.X.c);
	   }
	   if(goal != null && storedPos.containsKey("ball 0 1"))
	   {
		   debug(new Circle(storedPos.get("ball 1 1").position.c, 0.7));
		   ballDist = sphere.diff(goal.X.c, storedPos.get("ball 0 1").position.c);
	   }
	   if(storedPos.containsKey("ball 1 1") && storedPos.containsKey("ball 0 1"))
	   {
		   debug(new Circle(storedPos.get("ball 1 1").position.c, 0.7));
		   ballDist = sphere.diff(storedPos.get("ball 1 1").position.c, storedPos.get("ball 0 1").position.c);
	   }



	   if(ballDist > 33 && (ball != null || storedPos.containsKey("ball 0 1")))
	   {
		   status = "defending ";
		   defend();
		   return;
	   }

	   if(ball != null)
		   puckDist = sphere.diff(ball.X.c, self.X.c);
	   else if(storedPos.containsKey("ball 0 1"))
		   puckDist = sphere.diff(storedPos.get("ball 0 1").position.c, self.X.c);

	   // run out of energy, recover:
	   if(energy <= (Qmax* 0.1)){
		   status = "Recovering energy";
		   state = State.RECOVER;
		   return;
	   }

	   if(ball == null){ // ball not visible
		   status = "Looking for ball";
		   turn(false);
		   return;
	   }


	   double ballAngle = (sphere.warp(ball.X.c)).arg();

	   if(ballDist < 33){ // ball in goal range
		   status = "Going / turning to ball ";
		   goto_ball();
		   return;
	   }

	   if(goal == null){
		   status = "Looking for goal";
		   turn(true);
		   return;
	   }

	   double goalAngle = (sphere.warp(goal.X.c)).arg();
	   if (Math.abs(goalAngle) > 0.1){
		   status = "Turning to goal";
		   search_goal();
		   return;
	   }

	   double goalDist = sphere.diff(self.X.c, goal.X.c);
	   if (goalDist > maxShotDist){ // too far away: "dribble" in rough direction
		   status = "Passing ball away from goal";
		   action(0, 0, 0);
		   return;
	   }

	   if (ballDist > controlShootDist)
	   {
		   action(-0.01,0,0);
 	  	   target = ball.X.c;
 	  	   targetDist = sphere.diff(self.X.c, target);

	   }

	   /*if (goalDist < minShotDist){ // too close
		   status = "Getting out of violation range";
		   retreat();
		   return;
	   }*/

	   // else: ball in control range, goal close and in front:
	   // go for it!
	   status = "Aiming at goal"  + sphere.diff(self.X.c, ball.X.c);
	   precise_kick();
	   return;



   }

   // turns Unit with or without attracting the ball
   private void turn(boolean attract) {
	   double load = (attract)? -0.01 : 0;
	   action(load,-Fmax,Fmax);

   }

   /**
    * Wait until energy is recovered
    */
   private void recover(){
	   action(0,0,0);
	   if (energy >= (0.6* Qmax)){
		   state = State.THINK;
		   return;
	   }

   }

   /**
    * Move away from goal if too close
    */
   private void retreat(){
	   Puck goal = puck(BALL,self.t,1);
	   if (goal == null)
		   return;
	   target = goal.X.c;
	   targetDist = sphere.diff(self.X.c, goal.X.c);
	   action(-0.01, -Fmax, -Fmax);
	   debug(new Circle(goal.X.c, minShotDist/option("world.radius")), -1);

   }

   private void precise_kick() {
	   target = null;
	   Puck ball = puck(BALL,0,1);
	   Puck goal = puck(BALL,self.t,1);
	   if (ball == null || goal == null){
		   state = State.THINK;
		   return;
	   }
	   double goalDist = sphere.diff(self.X.c, goal.X.c);
	   double ballDist = sphere.diff(self.X.c, ball.X.c);
	   //double ballAngle = (sphere.warp(ball.X.c)).arg();

	   if (ballDist > controlDist ){
		   state = State.THINK;
		   return;
	   }

	   // TODO: needs to check if shot is possible at all by aiming
	   // (ball out of angle



	   Complex ballPos = sphere.warp(ball.X.c);

	   // Unit's center of charge:
	   Complex forceCenter = new Complex(0.5*option("unit.radius"), 0);

	   // direction of force from forceCenter to ball
	   Complex forceDirection = (Complex.sub(ballPos, forceCenter));
	   // from ball to goal:
	   Complex intendedForceDirection = (Complex.sub(sphere.warp(goal.X.c), ballPos));

	   // distance
	   double l = intendedForceDirection.abs() + ballDist - option("ball.radius");
	   forceDirection = forceDirection.norm(); /*immutable*/
	   intendedForceDirection = intendedForceDirection.norm(); /*immutable*/
	   Vector shotTarget = sphere.warp(Complex.add(forceCenter, Complex.mul(forceDirection, l)));
	   // how far would the goal be missed:
	   double missedBy = sphere.diff(goal.X.c, shotTarget);


	   if (missedBy < 2*option("ball.radius")){
		   action(0.15/20*goalDist, Fmax, Fmax);
		   state = State.THINK;
	   }
	   else{
		   double crossProd = Complex.vec(forceDirection, intendedForceDirection);
		   action(0,crossProd*Fmax, -crossProd*Fmax);
		   state = State.AIM;
	   }



	   debug(new Line(sphere.warp(forceCenter), shotTarget));
	   debug(new Line(ball.X.c, goal.X.c),-1);
}

   /**
    * Attract ball and wait for it to calm down
    */
	private void get_ball() {
		status = "Securing ball";
		Puck ball = puck(BALL,0,1);
		if (ball == null){  		// if no ball visible: start searching again
			return;
		}
		else{
			double ballDist = sphere.diff(ball.X.c, self.X.c);
			// difference in ball distance small => ball pos. stable
			if (ballDist > controlDist || Math.abs(ballDist - targetDist) < 0.01){
				state = State.THINK;
				return;
			}
			if(sphere.warp(ball.X.c).arg() < 0)
				action(-0.01, 33, 10);
			else
				action(-0.01, 10, 3);
			target = ball.X.c;
			targetDist = ballDist;
		}

	}


	/**
	 * Turns unit towards goal, keeps ball attracted
	 */
	private void search_goal() {
		Puck goal = puck(BALL,self.t,1);
		if (goal == null){
			return;
		}
		else{
			Complex x = sphere.warp(goal.X.c);
			target = goal.X.c;
			targetDist = sphere.diff(self.X.c, goal.X.c);
			double l = -x.arg();
     	  	double r =  x.arg();
			action(-0.01,l*Fmax,r*Fmax);
		}
	}

	/**
	 * Positions unit in front of ball
	 */
	private void goto_ball()
	{
		Puck ball = puck(BALL,0,1);
		if(ball == null){
			return;
		}

		// target position: in front of ball
		Complex ballWarp = sphere.warp(ball.X.c);
		Complex targetWarp = Complex.sub(ballWarp, (Complex.norm(ballWarp)).mul(touchingDist));


		// position in between ball and goal, if goal
		// visible and behind ball
		Puck goal = puck(BALL,self.t,1);
		if (goal != null){
			Complex goalWarp = sphere.warp(goal.X.c);
			Complex ballToGoal = Complex.sub(goalWarp, ballWarp);

			// goal behind ball
			if (sphere.diff(self.X.c, goal.X.c) >= sphere.diff(self.X.c, ball.X.c)){
				targetWarp = Complex.sub(ballWarp, (ballToGoal.norm()).mul(touchingDist));
			} else {// goal in between unit and ball
				double dist = Complex.vec(goalWarp,ballWarp) / ballWarp.abs();
				double securityDist = touchingDist +3;
				if (Math.abs(dist) <= securityDist){ // too close to goal
					double t = Complex.dot(goalWarp, ballWarp) / ballWarp.abs2();
					Complex closestPt = Complex.mul(ballWarp, t);
					targetWarp = Complex.add(goalWarp, Complex.sub(closestPt, goalWarp).norm().mul(securityDist));

				}
				debug(new Circle(goal.X.c, securityDist/option("world.radius")), -1);
			}

		}

		double thrust = Math.max(0.2,1-10*Math.abs(targetWarp.arg()));
		double l = -targetWarp.arg() + thrust;
  	  	double r =  targetWarp.arg() + thrust;
  	  	target = sphere.warp(targetWarp);
  	  	targetDist = sphere.diff(target, self.X.c);

  	  	/*if (targetDist <= controlDist)// got ball, keep it close
  	  	{
  	  		state = State.GET_BALL;
  	  		action(-0.01,0,0);
  	  		target = ball.X.c;
  	  		targetDist = sphere.diff(self.X.c, target);
		}*/
  	  	if(sphere.diff(self.X.c, ball.X.c) < 5 && (ball != null || storedPos.containsKey("ball 0 1")))
  	  	{
  	  		action(1, 0, 0);
  	  		return;
  	  	}
		else
		{
			action(0,l*Fmax,r*Fmax);
			return;
		}
	}

	private void defend()
	{
		Puck goal = puck(BALL,1,1);
		if (goal == null)  		// if no ball visible: start searching again
			return;

		Complex goalWarp = sphere.warp(goal.X.c);
		Complex targetWarp = Complex.sub(goalWarp, (Complex.norm(goalWarp)).mul(touchingDist));
		target = sphere.warp(targetWarp);
  	  	targetDist = sphere.diff(target, self.X.c);

		double thrust = Math.max(0.2,1-10*Math.abs(targetWarp.arg()));
		double l = -targetWarp.arg() + thrust;
  	  	double r =  targetWarp.arg() + thrust;

  	  		if(targetDist > 33)
  	  		{
		  		//state = State.GET_BALL;
  	  			action(0,l*Fmax,r*Fmax);
  	  		}
		else
		{
			status = "Looking for ball";
			action(0,-Fmax/2,Fmax/2);
 		   	return;
		}

	}

	/**
	 * this messages needs to be called at the beginning of the while(look()) loop.
	 * it stores all pucks in view in a map and updates positions of these pucks directly.
	 * unseen pucks from previous look()'s get a corrected position. the factor is determined
	 * by the change of relative positions of nodes.
	 */
	private void checkOrientation()
	{
		Puck tmpPuck;
		StoredNode storedPuck;
		// go through all nodes and update stored data
		for (Iterator<String> iterator = storedPos.keySet().iterator(); iterator.hasNext();) {
			String key = (String) iterator.next();
			storedPos.get(key).active = false;
		}
		for ( int i = 0; i < pucks.length; i++ )
		{
			tmpPuck = pucks[i];
			if ( tmpPuck.X != null &&tmpPuck.id.startsWith("node")  )
			{
				if ( !storedPos.containsKey(tmpPuck.id) )
				{
					storedPos.put(tmpPuck.id, new StoredNode(tmpPuck) );
				}
				else
				{
					storedPuck = storedPos.get(tmpPuck.id);
					storedPuck.updatePosition(tmpPuck.X);
				}
			}
			else if ( tmpPuck.X != null )
			{
				if ( !storedPos.containsKey(tmpPuck.id) )
				{
				storedPos.put(tmpPuck.id, new StoredNode(tmpPuck) );
				}
				else
				{
					storedPuck = storedPos.get(tmpPuck.id);
					storedPuck.updatePosition(tmpPuck.X);
				}
			}

		}
		/*
		now all pucks in view are added to the list or updated,
		unseen pucks are inactive
		get a node that was active last round and is still active
		*/
		StoredNode tmpNode = null;
		for (Iterator<String> iterator = storedPos.keySet().iterator(); tmpNode == null && iterator.hasNext();) {
			String key = (String) iterator.next();
			if ( storedPos.get(key).active && storedPos.get(key).id.startsWith("node") )
				tmpNode = storedPos.get(key);
		}
		//debug(new XVector(tmpNode.previous.c),199);
		//debug(new XVector(tmpNode.position.c));
		Matrix move = Matrix.mul(Matrix.inv(tmpNode.position), tmpNode.previous );
		//System.out.println("move: "+move+" length: "+move.c.abs());

		// update stored positions of inactive pucks
		for (Iterator<String> iterator = storedPos.keySet().iterator(); iterator.hasNext();)
		{
			String key = (String) iterator.next();
			storedPos.get(key).correctPosition(move);
		}
	}
}
