
// *************************************************************************
//   Cosmic Friends (Hoverball Team)
//   Version 1.4
// *************************************************************************
//   August 2007, Stefan Bornhofen                           www.hoverball.de
// *************************************************************************


/*
   Aufruf als Application:

     java hoverball.team.CosmicFriends                    - öffnet die Cosmic Friends
     java hoverball.team.CosmicFriends name               - öffnet die Cosmic Friends unter dem Namen "name"
     java hoverball.team.CosmicFriends name color         - öffnet die Cosmic Friends unter dem Namen "name"
                                             mit der Farbe "color" (der Form "RRGGBB")
   Aufruf mit Constructor:

     CosmicFriends()                       - öffnet die Cosmic Friends
     CosmicFriends(String name)            - öffnet die Cosmic Friends unter dem Namen "name"
     CosmicFriends(String name, int color) - öffnet die Cosmic Friends unter dem Namen "name"
                                             mit der Farbe color
*/

package hoverball.team.CosmicFriends_;

import hoverball.*;
import hoverball.debug.*;
import hoverball.math.*;


/*************************************************************************
  Team: CosmicFriends
  Strategie:
    - Moondancer hält sich immer in der Verbindungsgerade zwischen gegnerischem
      Team Ball (flop) und Ball auf. Er ist sozusagen der Torhüter.
    - Mad Dragon und Angelfire versuchen, Tore zu schiessen. Der dichtere Puck führt
      zum Ball und schiesst ihn Richtung Tor, der andere Spieler bewegt sich hinter das Tor und lauert.
    Dabei vermeiden alle Spieler eventuelle Kollisionen mit Tor und Flop,
    welche zu Energieverlust führen.
*************************************************************************/


public class CosmicFriends extends Team
{
  // ---- Application

  public static void main (String[] args)
  {
    Application.setBounds();

    if (args.length == 0) new CosmicFriends().show();
    else if (args.length == 1) new CosmicFriends(args[0]).show();
    else new CosmicFriends(args[0],Integer.parseInt(args[1],16)).show();
  }

  // ---- Constructor

  public CosmicFriends () { this(null); }
  public CosmicFriends (String team) { this(team,0xFFCC00); }
  public CosmicFriends(int color) { this(null,color); }
  public CosmicFriends (String team, int color)
  {
    super((team == null || team == "")? "Cosmic Friends" : team,
       "1.4 (August 2007)","Stefan Bornhofen","Let's win the Hoverball Competition 2007!");

    // main.setBackground(new java.awt.Color(0xFFBB33));

    add(new CosmicFriend(null,"Moondancer",color,CosmicFriend.MOONDANCER));
    add(new CosmicFriend(null,"Mad Dragon",color,CosmicFriend.MADDRAGON));
    add(new CosmicFriend(null,"Angelfire", color,CosmicFriend.ANGELFIRE));
  }
}

// *************************************************************************
// Spieler: CosmicFriend
// *************************************************************************


class CosmicFriend extends Unit
{
  private final String name;                // mein Name
  private final int identity;               // meine Nummer (MOONDANCER = 0, MADDRAGON = 1, ANGELFIRE = 2)
  private int nbOpponents = 0;        // zaehlt entdeckte Gegnerunits, 3 maximal (fuer regulaeres Spiel)
  private boolean pucksLocated = false;    // ich kann erst losspielen, wenn ich alle Pucks geortet habe
  // aktuellen Daten
  private Puck ball = null;               // der Ball
  private Puck goal = null;               // das eigene Tor
  private Puck flop = null;               // das gegnerische Tor
  private Puck friend[] = new Puck[3];    // mein Team (ich eingeschlossen)
  private Puck opponent[] = new Puck[3];  // Gegnerteam
  private Puck node[] = new Puck[6];      // die Knoten

  // die Knoten, die ich das letzte Mal gesehen habe
  private Puck node_[] = new Puck[6];      // die alten Knoten

  private double P = 0; // Aktion: Polarisation
  private double L = 0; // Aktion: linker Motor
  private double R = 0; // Aktion: rechter Motor

  private final double DANGERRADIUS = 20;   // Abstand zum Teamball von Moondancer als Torhueter und von Mad Dragon / Angelfire auf Lauerstellung
  private final double AVOIDRADIUS = 15;   // Ab dieser Distanz muesses Teamballs umfahren werden

  final static int MOONDANCER = 0;
  final static int MADDRAGON = 1;
  final static int ANGELFIRE = 2;

  String reference = ""; // der Referenzknoten

  // ---- Constructor ---------------------------------------------------------

  CosmicFriend (String team, String name, int color, int identity) {
    super(team,name,color);
    this.name = name;
    this.identity = identity;
    // main.setBackground(new java.awt.Color(0xFFBB33));
  }

  // ---- Loop ----------------------------------------------------------------

  private Sphere sphere     = null;         // die Kugeloberfläche
  private double sphere_rad = 0;            // Kugelradius
  private double charge_max = 0;            // max. positive Polarisation
  private double charge_min = 0;            // max. positive Polarisation
  private double engine_min = 0;            // min negativer Motorenwert
  private double engine_max = 0;            // max positiver Motorenwert
  private double distanceBallUnit = 0;      // minimaler Abstand Ball-Unit
  private double radius = 0;                // eigener Radius

  public void loop () {
    sphere_rad = option("world.radius");
    sphere = new Sphere(sphere_rad);
    charge_min = option("unit.charge.min");
    charge_max = option("unit.charge.max");
    engine_min = option("unit.engine.min");
    engine_max = option("unit.engine.max");
    radius = option("unit.radius");
    distanceBallUnit = radius + option("ball.radius");

    while (look())  // loop structure:  while (look()) { ...; action(...); }
    	    {
      P = 0;  // Aktionen initialisieren
      L = 0;
      R = 0;
      String actionMessage = "";

      // temporäre Variablen zur Berechnung
      Complex c1 = null;
      Complex c2 = null;
      Complex c3 = null;
      Complex c4 = null;
      Vector v1 = null;

      detect(pucks);

      if (!pucksLocated) { // noch nicht alles gesehen, dreh Dich
    	rotate();
        actionMessage = "detect";
      }
      else {
        switch (identity) {
          case MOONDANCER: // Moondancer ist der Torhueter
        	  if (see(flop))
        	  {
        		  c1 = warp(flop.X);
        		  c2 = warp(ball.X);
        		  c3 = Complex.sub(c2,c1).mul(Complex.expi(Math.PI/32));
        		  c4 = Complex.add(c1,Complex.mul(c3,DANGERRADIUS/c3.abs()));
        		  v1 = sphere.warp(c4);
        		  if ((distanceTo(ball)<3*distanceBallUnit) || (distance(ball,flop)<DANGERRADIUS)) { // Gefahr, weg mit dem Ball!
        			  clear(ball.X.c);
        			  actionMessage = "clear";
        		  }
        		  else if (sphere.diff(v1,self.X.c)>0.2*radius) { // stell Dich in die Verbindunglinie ball - flop
        			  goTo(v1,50);
        			  actionMessage = "position";
        		  }
        		  else { // dreh Dich und beobachte
        			  rotate();
        			  actionMessage = "rotate";
        		  }
        		  debug (new Circle(flop.X.c,DANGERRADIUS/sphere_rad));
        		  debug (new Line(new Vector(0,0,0),v1));
        	  }
        	  else { // dreh Dich und such den flop
        			  rotate();
        			  actionMessage = "search";
        	  }
          break;
          case MADDRAGON: // Mad Dragon &
          case ANGELFIRE: // Angelfire gehen zum Ball
            Puck partner = null;
            if (identity == MADDRAGON) partner = friend[ANGELFIRE]; else partner = friend[MADDRAGON];
            if (partner!=null) {
              if (distanceTo(ball)<distance(partner,ball)) { // ich bin dichter als der andere -> gehe zum Ball
                shoot(ball.X.c,goal.X.c,0.01);
                actionMessage = "shoot";
              }
              else { // gehe auf die gegenüberliegende Seite und warte auf Torschuss
                c1 = warp(goal.X);
                c2 = warp(ball.X);
                c3 = Complex.sub(c1,c2);
                c4 = Complex.add(c1,Complex.mul(c3,DANGERRADIUS/c3.abs()));
                v1 = sphere.warp(c4);
                if (sphere.diff(v1,self.X.c)>4*radius) {
                  goTo(v1,50);
                  actionMessage = "lurk";
                }
                else {
                  rotate();
                  actionMessage = "rotate";
                }
                debug (new Line(new Vector(0,0,0),v1));
              }
            }
          break;
        }
      }
      action(P,L,R,""+identity); // abschicken der Aktionen
       // ----------------- Debugging ------------------------
      debug (new Text(null,name,1,0));
      debug (new Text(null,actionMessage,2,0));
      debug (new Text(null,"reference: "+reference,3,0));
      // ------------------- Remember -----------------------
     } // look
  } // loop

  /*************************************************************************
   * function clear: Vector
   ************************************************************************/

  private void clear(Vector v) {
    if (distanceTo(v)<1.5* distanceBallUnit)
      P = charge_max;
    else {
      P = 0;
      goTo(v,1);
    }
  }

 /*************************************************************************
  * function shoot: Vector, Vector, double
  ************************************************************************/

  private void shoot(Vector object, Vector target, double angle) {
  // Schiesst ein Object in eine gewisse Richtung target,
  // wenn der Wingel angle erreicht ist.
    Complex oc = sphere.warp(object);
    Complex tc = sphere.warp(target);
    if ((distanceTo(object)>2*distanceBallUnit))
    {
        P = 0;
        goTo(object,1);
    }
    else if ((Math.abs(oc.arg()-tc.arg())>angle))
    {
    	boolean opponentNear = false;
    	for (int i=0; i<nbOpponents; i++) if (distance(object,opponent[i].X.c)<1.5*distanceBallUnit) opponentNear = true;
    	if (opponentNear) P = charge_max; // clear is there is an opponent
    		else P = charge_min/100.0;
        turnTo(goal.X.c);
    }
    else
    {
    	P = charge_max;
    }
  }

  /**************************************************************************
   * function avoid: Vector
   *************************************************************************/

  private Vector avoid (Vector v) {
  // Man muss goal und flop vermeiden, wenn man auf v zuführt
	Vector x = avoid(v,goal);
    if (see(flop)) x = avoid(x,flop);
    return x;
  }

  private Vector avoid(Vector v, Puck p) {
  // Vermeidet den Puck p
    if (p!=null) if (distanceTo(p)<AVOIDRADIUS) {
    // Hindernis umfahren
      Complex c1 = warp(p.X);
      Complex c2 = sphere.warp(v);
      double ang = triangle(new Complex(),c1,c2);
      if (Math.abs(ang)>Math.PI/2)
        return sphere.warp(Complex.add(Complex.sub(c2,c1).mul(Complex.expi(-ang/2)),c1));
    }
    return v;
  }

  /**************************************************************************
   * function goTo: Vector
   *************************************************************************/

  private void goTo(Vector v, double brake) {
  // der Spieler führt zum Vector v und vermeidet dabei goal und flop

    Vector v_ = avoid(v);
    double angle = angle(v_)/Math.PI; //angle in [-1,1]
    if (Math.abs(angle)>0.05) turnTo(v_); // Ziel nicht vor mir -> dreh dich
    else { // geradeaus volle Power
      double d = distanceTo(v_);
  	  L = engine_max*d*d/brake;
   	  R = engine_max*d*d/brake;
     }
  }

  /**************************************************************************
   * function turnTo: Vector
   *************************************************************************/

  private void turnTo(Vector v) {
    double angle = angle(v)/Math.PI; //angle in [-1,1]
    L = -100*engine_max*angle;
    R = -L;
  }

  /**************************************************************************
   * function rotate: no argument
   *************************************************************************/
  private void rotate() {
      L = engine_max;
      R = engine_min;
  }

  // ************************ Weitere Hilfsfunktionen *****************************

  private Complex warp (Matrix M) {
  // 3D-Matrix -> 2D komplexer Vektor
    Vector v = sphere.warp(M);
    return new Complex(v.x,v.y);
  }

  private double angle(Vector v) {
  // berechnet den Winkel zum Vector v
    if (v!=null) return new Complex(v).arg(); else return 0;
  }

  private double triangle(Complex a,Complex v,Complex b) {
  // berechnet den Winkel zwischen Vector v und a und b, v am Winkel
    if ((v!=null) && (a!=null) && (b!=null)) {
      Complex va = Complex.sub(a,v);
      Complex vb = Complex.sub(b,v);
      double x = vb.arg()-va.arg();
      if (x>Math.PI) x-=2*Math.PI;
      else if (x<-Math.PI) x+=2*Math.PI;
      return x;
    }
    else return 2*Math.PI;
  }

  private double distance(Vector v1, Vector v2) {
  // berechnet die Distanz zwischen den Vektoren v1 und v2
    if ((v1!=null) && (v2!=null)) return sphere.diff(v1,v2); else return sphere.rad*7;
  }

  private double distance(Puck p1,Puck p2) {
  // berechnet die Distanz zwischen den Pucks p1 und p2
    if ((p1!=null) && (p2!=null)) return distance(p1.X.c,p2.X.c); else return sphere.rad*7;
  }

  private double distanceTo(Puck p) {
  // berechnet die Distanz zwischen mir und Puck p
    if (p!=null) return distance(self.X.c,p.X.c);
    else return sphere.rad*7;
  }

  private double distanceTo(Vector v) {
  // berechnet die Distanz zwischen mir und Vector v
    return distance(self.X.c,v);
  }

  private boolean see(Puck p)
  {
	  if (p==null) return false;
	  if (p.X==null) return false;
	  return true;
  }

  private void detect(Puck[] pucks) {
    // Sucht die Pucks aus dem Array pucks zusammen und speichert alles schön ordentlich
    // in den Variablen ball, goal, flop, friend, opponent, node

    // Korrigiere die bekannten Daten anhand eines Knotens
	Matrix m = null;
	for (int i=0; i<6; ++i)
		if ((m==null) && (node_[i]!=null)) if (see(puck(NODE,0,i+1))) {
			reference = "node"+(i+1);
			m = Matrix.inv(node_[i].X).mul(puck(NODE,0,i+1).X);
			break;
		}

    if (m!=null) // muesste eigentlich immer der Fall sein, wenn unit.vision = 0.66
    {
        if (see(ball)) ball.X = ball.X.mul(m); /*immutable*/
        if (see(goal)) goal.X = goal.X.mul(m); /*immutable*/
        if (see(flop)) flop.X = flop.X.mul(m); /*immutable*/
        for (int k=0; k<3; ++k) if (see(friend[k])) friend[k].X = friend[k].X.mul(m); /*immutable*/
        for (int k=0; k<nbOpponents; ++k) if (see(opponent[k])) opponent[k].X = opponent[k].X.mul(m); /*immutable*/
        for (int k=0; k<6; ++k) if (see(node[k])) node[k].X = node[k].X.mul(m); /*immutable*/
    }

    // setze die alten Knoten auf null
    for (int i=0; i<6; i++) node_[i] = null;

    // ueberschreibe mit eventuellen neuen Daten
    for (int i=0; i<pucks.length; ++i) {
    	if (pucks[i].X != null) // ich muss sie auch sehen, nicht nur hoeren
    	{
    		if (pucks[i].equals(BALL))
    		{ // ein Ball ist's
    			if (pucks[i].t==0) ball = new Puck(pucks[i]);
    			else if (pucks[i].t==self.t) goal = new Puck(pucks[i]); // es ist ein Tor
    			else flop = new Puck(pucks[i]); // es ist ein Flop
    		}
    		else if (pucks[i].equals(UNIT))
    		{ // eine Unit ist's
      			if (pucks[i].t==self.t) {
      				if (pucks[i].message.length()>0)
      					friend[Integer.parseInt(pucks[i].message)] = new Puck(pucks[i]); // ... noch dazu aus meinem Team
      			}
      			else
      			{ // ... vom Gegnerteam
      				boolean found = false;
      				for (int j=0; j<nbOpponents; ++j) if (opponent[j].id.equals(pucks[i].id))
      				{
      					opponent[j] = new Puck(pucks[i]);
      					found = true;
      				}
      				if ((!found) && (nbOpponents<3)) opponent[nbOpponents++] = new Puck(pucks[i]);
      			}
    		}
    		else if (pucks[i].equals(NODE)) // ein Knoten ist's
    		{
    			node[pucks[i].n-1] = new Puck(pucks[i]);
    			node_[pucks[i].n-1] = new Puck(pucks[i]);
    		}
    	}
    }
    // vielleicht ist pucksLocated jetzt true?
    if (!pucksLocated) pucksLocated = locate();
}

  boolean locate() {
  // der Spieler kann erst dann loslegen, wenn er gewisse Pucks lokalisiert hat
    if (ball==null) return false;
    if (goal==null) return false;
    for (int i=0; i<3; i++) if (friend[i]==null) return false;
    for (int i=0;i<6; i++) if (node[i]==null) return false;
    return true;
  }
}
