/* Hoverball Competition 2007
 *
 * Team   : Duck Tales
 * Author : Sven Wunderlich, Louis Lauser
 */

package hoverball.team.DuckTales_;

import hoverball.*;
import hoverball.debug.*;
import hoverball.math.*;

public class DuckTales extends Team
{
  // ---- Application

  public void close () { super.close(); }

  public static void main (String[] args)
  {
    Application.setBounds();

    new DuckTales("Duck Tales").show();
  }

  // ---- Constructor

  public DuckTales() { this(null); }
  public DuckTales(String team) { this(team,-1); }
  public DuckTales(int color) { this(null,color); }
  public DuckTales(String team, int color)
  {
    super((team == null || team == "")? "Duck Tales" : team,
       null,"Sven Wunderlich, Louis Lauser",null);

	add(new Tick("Tick",(color<0)? 0x00FF00 : color));
	add(new Tick("Trick",(color<0)? 0x00FF00 : color));
	add(new Track((color<0)? 0x0044FF : color));
  }
}

// ---------------------------------------------------------------

class Track extends Unit
{
    public Track(int color) {
	    super("Duck Tales","Track",color);
    }

	Puck ball = null;
	Puck goal = null;

	Matrix pov = null;
	Matrix ballCoords = null;
	Matrix goalCoords = null;
	Matrix dest = null;


   public void loop()
   {
		Sphere sphere = new Sphere(option("world.radius"));  // Parameter auslesen...
		double Qmax = option("unit.charge.max");
		double Fmax = option("unit.engine.max");


		while (look())
		{
			pov = getPOV();

			ball = puck(BALL,0,1);       // das ist der Ball
			goal = puck(BALL,3-self.t,1);  // das ist das "Tor"

			getHiddenCoords ();

			if (ballCoords == null || goalCoords == null) action(0,-Fmax,Fmax);     // sieht keinen Ball? drehe!
			else {                                      // sieht Ball:
//				Vector a = Vector.vec(goalCoords.c,ballCoords.X.c); // berechne Drehachse
//				Matrix X = Matrix.rot(a,(self.r+ball.r+self.r/10)/sphere.rad).mulL(ball.X);
//				Complex x = sphere.warp(X.c);            // X ist die Schussposition

				Complex b = sphere.warp(getRelativeCoords(ballCoords).c);
				Complex t = sphere.warp(getRelativeCoords(goalCoords).c);

				double Abstand = 14;
				Complex tb = Complex.sub(b,t);
				double ra = Abstand/tb.abs();

				Complex x = Complex.mul(tb,ra).add(t); // x = ra*(b-t)+t;

				double rb = Abstand/x.abs();
				Complex x_back = Complex.mul(x,rb);
				Complex y = Complex.sub(x,x_back);
				Complex ty = Complex.sub(y,t);
				double rc = 4.4/ty.abs();
				Complex x2 = Complex.mul(ty,rc);


				if (ty.abs()<3.2) {
					x = Complex.add(y,x2);
				}

				double entfernung = t.abs();
				double rd = entfernung/tb.abs();
				Complex tb_punkt = Complex.mul(tb,rd).add(t);

				double v_factor_b = 1;
				if ((tb_punkt.abs()<3) && (Complex.add(t,b).abs() != (t.abs() + b.abs()))) {
					x = b;
					v_factor_b = 22*(1/b.abs());
				}

				double bremsen;
				if((x.abs()<4.8) && (x.abs()>2) && (x!=x2) && (x!=b)) {
					bremsen = -100;
				} else {
					bremsen = 1;
				}

				double v_factor;
				if((x.abs()<3) && (x!=x2) && (x!=b)) {
					v_factor = x.abs()/10;
				} else {
					v_factor = 1;
				}

				double q;
				if(b.abs()<6) {
					q=1;
				} else {
					q=0;
				}

				Vector Sx = sphere.warp(x);
				Vector Sy = sphere.warp(y);
				Vector Sx2 = sphere.warp(y);
//				System.out.println(""+Sx);
				debug(new Line(self.X.c,Sx),-1); //new hoverball.debug.X(Sx));
//				debug(new Line(self.X.c,Sy),-1);
//				debug(new Line(self.X.c,Sx2),-1);

				double l = -x.arg() + Math.max(0.2,1-10*Math.abs(x.arg())); // drehe und
				double r =  x.arg() + Math.max(0.2,1-10*Math.abs(x.arg())); // fahre zu X
//				double q = (x.abs() < 1)? 0.5 : 0;  // falls auf X: Schuss!

				action(q*Qmax,l*Fmax*bremsen*v_factor*v_factor_b,r*Fmax*bremsen*v_factor*v_factor_b);


//				dest = new Matrix(X);
			}

			try {
				debug (new Text(pov.c, "POV", 2), -1);
				debug (new Text(getRelativeCoords(goalCoords).c, "GOAL", 2), -1);
//				debug (new Circle (dest.c, Math.PI/100));
				debug (new Line (getRelativeCoords(ballCoords).c, getRelativeCoords(goalCoords).c), -1);
			} catch (NullPointerException err) {};
		}

   }

	private void getHiddenCoords() {
		if (ball != null) {
			ballCoords = getAbsoluteCoords(ball.X);
		}
		if (goal != null) {
			goalCoords = getAbsoluteCoords(goal.X);
		}
	}

	private Matrix getRelativeCoords(Matrix m) {
		return new Matrix(pov).mulL(m);
	}

	private Matrix getAbsoluteCoords(Matrix m) {
		return Matrix.inv(this.pov).mulL(m);
	}

	private Matrix getPOV() {
		Matrix pov;
		Puck node = puck (NODE, 0, 1);
		if (node != null) pov = node.X;
		else {
			node = puck (NODE, 0, 6);
			pov = Matrix.neg(node.X);
			pov = new Matrix(pov.a,pov.b.neg(),pov.c); /*immutable*/  // pov.b = pov.b.neg();
		};
		return pov;
	}
}


class Tick extends Unit
{
   public Tick(String name, int color) {
	   super("Duck Tales",name,color);
	   state = THINK;
   }
   // available states of Aragorn:
   final int THINK = 1;
   final int AIM = 2;
   final int GET_BALL = 3;
   final int RECOVER = 4;

//   public enum State {THINK, AIM, GET_BALL, RECOVER }

   private int state;
   private double Qmax, Fmax;
   private Sphere sphere;
   private Vector target;
   private double targetDist;
   private String status;
   private double touchingDist;
   private double controlDist;
   private double controlShootDist;
   private double controlAngle;
   private double maxShotDist;
   private double minShotDist;


   public void loop()
   {
	   //Initialization:
	  sphere = new Sphere(option("world.radius"));
      Qmax = option("unit.charge.max");
      Fmax = option("unit.engine.max");
      touchingDist = option("unit.radius")+2*option("ball.radius");
      controlDist = touchingDist + 4;
      controlShootDist = touchingDist + 1;
      controlAngle = Math.PI/6; // +-30°: 60° control range
      //maxShotDist = option("world.radius")*Math.PI*0.5;
      maxShotDist = 45;
      minShotDist = 20;

      while(look()){

    	  switch (state) {
			case GET_BALL:
				get_ball();
			break;
			case AIM:
				precise_kick();
				break;
			case RECOVER:
				recover();
				break;
			default:
				think();
				break;
		 }


    	  targetDist = Math.floor(targetDist * 100.0) / 100.0;
    	  String energyStr = Double.toString(Math.floor(energy*Qmax*100.0) /100.0);
    	  debug(new Text(self.X.c, "Aragorn \n E="+energyStr+"\n status: "+status+"\n d="+targetDist , 1), -1);
    	  if (target != null)
    		  debug(new Line(self.X.c, target));


      } // end of while(look())

   } // end of loop()

   private void think(){
	   Puck ball = puck(BALL,0,1);
	   Puck goal = puck(BALL,self.t,1);
	   target = null;
	   targetDist = 0;
	   status = "";

	   // run out of energy, recover:
	   if (energy <= (Qmax* 0.1)){
		   status = "Recovering energy";
		   state = RECOVER;
		   return;
	   }

	   if (ball == null){ // ball not visible
		   status = "Looking for ball";
		   turn(false);
		   return;
	   }

	   double ballDist = sphere.diff(self.X.c, ball.X.c);
	   double ballAngle = (sphere.warp(ball.X.c)).arg();

	   if (ballDist > controlDist || Math.abs(ballAngle) > controlAngle){ // ball out of range
		   status = "Going / turning to ball";
		   goto_ball();
		   return;
	   }


	   if (goal == null){
		   status = "Looking for goal";
		   turn(true);
		   return;
	   }

	   double goalAngle = (sphere.warp(goal.X.c)).arg();
	   if (Math.abs(goalAngle) > 0.1){
		   status = "Turning to goal";
		   search_goal();
		   return;
	   }

	   double goalDist = sphere.diff(self.X.c, goal.X.c);
	   if (goalDist > maxShotDist){ // too far away: "dribble" in rough direction
		   status = "Passing ball closer to goal";
		   action(0.1, 0, 0);
		   return;
	   }

	   if (ballDist > controlShootDist){
		   action(-0.01,0,0);
 	  	   target = ball.X.c;
 	  	   targetDist = sphere.diff(self.X.c, target);

	   }

	   if (goalDist < minShotDist){ // too close
		   status = "Getting out of violation range";
		   retreat();
		   return;
	   }

	   // else: ball in control range, goal close and in front:
	   // go for it!
	   status = "Aiming at goal";
	   precise_kick();
	   return;



   }

   // turns Unit with or without attracting the ball
   private void turn(boolean attract) {
	   double load = (attract)? -0.01 : 0;
	   action(load,-Fmax,Fmax);

   }

   /**
    * Wait until energy is recovered
    */
   private void recover(){
	   action(0,0,0);
	   if (energy >= (0.6* Qmax)){
		   state = THINK;
		   return;
	   }

   }

   /**
    * Move away from goal if too close
    */
   private void retreat(){
	   Puck goal = puck(BALL,self.t,1);
	   if (goal == null)
		   return;
	   target = goal.X.c;
	   targetDist = sphere.diff(self.X.c, goal.X.c);
	   action(-0.01, -Fmax, -Fmax);
	   debug(new Circle(goal.X.c, minShotDist/option("world.radius")), -1);

   }

   private void precise_kick() {
	   target = null;
	   Puck ball = puck(BALL,0,1);
	   Puck goal = puck(BALL,self.t,1);
	   if (ball == null || goal == null){
		   state = THINK;
		   return;
	   }
	   double goalDist = sphere.diff(self.X.c, goal.X.c);
	   double ballDist = sphere.diff(self.X.c, ball.X.c);
	   double ballAngle = (sphere.warp(ball.X.c)).arg();

	   if (ballDist > controlDist ){
		   state = THINK;
		   return;
	   }

	   // TODO: needs to check if shot is possible at all by aiming
	   // (ball out of angle



	   Complex ballPos = sphere.warp(ball.X.c);

	   // Unit's center of charge:
	   Complex forceCenter = new Complex(0.5*option("unit.radius"), 0);

	   // direction of force from forceCenter to ball
	   Complex forceDirection = (Complex.sub(ballPos, forceCenter));
	   // from ball to goal:
	   Complex intendedForceDirection = (Complex.sub(sphere.warp(goal.X.c), ballPos));

	   // distance
	   double l = intendedForceDirection.abs() + ballDist - option("ball.radius");
	   forceDirection = forceDirection.norm(); /*immutable*/
	   intendedForceDirection = intendedForceDirection.norm(); /*immutable*/
	   Vector shotTarget = sphere.warp(Complex.add(forceCenter, Complex.mul(forceDirection, l)));
	   // how far would the goal be missed:
	   double missedBy = sphere.diff(goal.X.c, shotTarget);


	   if (missedBy < 2*option("ball.radius")){
		   action(0.15/20*goalDist, Fmax, Fmax);
		   state = THINK;
	   }
	   else{
		   double crossProd = Complex.vec(forceDirection, intendedForceDirection);
		   action(0,crossProd*Fmax, -crossProd*Fmax);
		   state = AIM;
	   }



	   debug(new Line(sphere.warp(forceCenter), shotTarget));
	   debug(new Line(ball.X.c, goal.X.c),-1);
}

   /**
    * Attract ball and wait for it to calm down
    */
	private void get_ball() {
		status = "Securing ball";
		Puck ball = puck(BALL,0,1);
		if (ball == null){  		// if no ball visible: start searching again
			return;
		}
		else{
			double ballDist = sphere.diff(ball.X.c, self.X.c);
			// difference in ball distance small => ball pos. stable
			if (ballDist > controlDist || Math.abs(ballDist - targetDist) < 0.01){
				state = THINK;
				return;
			}
			action(-0.01,0,0);
			target = ball.X.c;
			targetDist = ballDist;
		}

	}


	/**
	 * Turns unit towards goal, keeps ball attracted
	 */
	private void search_goal() {
		Puck goal = puck(BALL,self.t,1);
		if (goal == null){
			return;
		}
		else{
			Complex x = sphere.warp(goal.X.c);
			target = goal.X.c;
			targetDist = sphere.diff(self.X.c, goal.X.c);
			double l = -x.arg();
     	  	double r =  x.arg();
			action(-0.01,l*Fmax,r*Fmax);
		}
	}

	/**
	 * Positions unit in front of ball
	 */
	private void goto_ball() {
		Puck ball = puck(BALL,0,1);
		if (ball == null){
			return;
		}

		// target position: in front of ball
		Complex ballWarp = sphere.warp(ball.X.c);
		Complex targetWarp = Complex.sub(ballWarp, (Complex.norm(ballWarp)).mul(touchingDist));


		// position in between ball and goal, if goal
		// visible and behind ball
		Puck goal = puck(BALL,self.t,1);
		if (goal != null){
			Complex goalWarp = sphere.warp(goal.X.c);
			Complex ballToGoal = Complex.sub(goalWarp, ballWarp);

			// goal behind ball
			if (sphere.diff(self.X.c, goal.X.c) >= sphere.diff(self.X.c, ball.X.c)){
				targetWarp = Complex.sub(ballWarp, (ballToGoal.norm()).mul(touchingDist));
			} else {// goal in between unit and ball
				double dist = Complex.vec(goalWarp,ballWarp) / ballWarp.abs();
				double securityDist = touchingDist +3;
				if (Math.abs(dist) <= securityDist){ // too close to goal
					double t = Complex.dot(goalWarp, ballWarp) / ballWarp.abs2();
					Complex closestPt = Complex.mul(ballWarp, t);
					targetWarp = Complex.add(goalWarp, Complex.sub(closestPt, goalWarp).norm().mul(securityDist));

				}
				debug(new Circle(goal.X.c, securityDist/option("world.radius")), -1);
			}

		}

		double thrust = Math.max(0.2,1-10*Math.abs(targetWarp.arg()));
		double l = -targetWarp.arg() + thrust;
  	  	double r =  targetWarp.arg() + thrust;
  	  	target = sphere.warp(targetWarp);
  	  	targetDist = sphere.diff(target, self.X.c);

  	  	if (targetDist <= controlDist){ // got ball, keep it close
  	  		state = GET_BALL;
  	  		action(-0.01,0,0);
  	  		target = ball.X.c;
  	  		targetDist = sphere.diff(self.X.c, target);
		}
		else{
			action(0,l*Fmax,r*Fmax);

		}


	}
}

class Trick extends Unit
{
  private final String name;              // mein Name
  private final int identity;             // meine Nummer (MOONDANCER = 0, MADDRAGON = 1, ANGELFIRE = 2)
  private int nbOpponents = 0;            // zaehlt entdeckte Gegnerunits, 3 maximal (fuer regulaeres Spiel)
  private boolean pucksLocated = false;   // ich kann erst losspielen, wenn ich alle Pucks geortet habe
  // aktuellen Daten
  private Puck ball = null;               // der Ball
  private Puck goal = null;               // das eigene Tor
  private Puck flop = null;               // das gegnerische Tor
  private Puck friend[] = new Puck[3];    // mein Team (ich eingeschlossen)
  private Puck opponent[] = new Puck[3];  // Gegnerteam
  private Puck node[] = new Puck[6];      // die Knoten
  // die Knoten, die ich das letzte Mal gesehen habe
  private Puck node_[] = new Puck[6];     // die alten Knoten

  private double P = 0; // Aktion: Polarisation
  private double L = 0; // Aktion: linker Motor
  private double R = 0; // Aktion: rechter Motor

  private final double DANGERRADIUS = 20; // Abstand zum Teamball von Moondancer als Torhueter und von Mad Dragon / Angelfire auf Lauerstellung
  private final double AVOIDRADIUS = 15;  // Ab dieser Distanz muesses Teamballs umfahren werden

  final static int MOONDANCER = 0;
  final static int MADDRAGON = 1;
  final static int ANGELFIRE = 2;

  String reference = ""; // der Referenzknoten

  // ---- Constructor ---------------------------------------------------------

  public Trick (String team, String name, int color, int identity) {
    super(team,name,color);
    this.name = name;
    this.identity = identity;
    main.setBackground(new java.awt.Color(0xFFBB33));
  }

  // ---- Loop ----------------------------------------------------------------

  private Sphere sphere     = null;         // Kugeloberflche
  private double sphere_rad = 0;            // Kugelradius
  private double charge_max = 0;            // max. positive Polarisation
  private double charge_min = 0;            // max. positive Polarisation
  private double engine_min = 0;            // min negativer Motorenwert
  private double engine_max = 0;            // max. positiver Motorenwert
  private double distanceBallUnit = 0;      // minimaler Abstand Ball-Unit
  private double radius = 0;                // eigener Radius

  public void loop () {
    sphere_rad = option("world.radius");
    sphere = new Sphere(sphere_rad);
    charge_min = option("unit.charge.min");
    charge_max = option("unit.charge.max");
    engine_min = option("unit.engine.min");
    engine_max = option("unit.engine.max");
    radius = option("unit.radius");
    distanceBallUnit = radius + option("ball.radius");

    while (look())  // loop structure:  while (look()) { ...; action(...); }
    	    {
      P = 0;  // Aktionen initialisieren
      L = 0;
      R = 0;
      String actionMessage = "";

      // temporre Variablen zur Berechnung
      Complex c1 = null;
      Complex c2 = null;
      Complex c3 = null;
      Complex c4 = null;
      Vector v1 = null;

      detect(pucks);

      if (!pucksLocated) { // noch nicht alles gesehen, dreh Dich
    	rotate();
        actionMessage = "detect";
      }
      else {
        switch (identity) {
          case MOONDANCER: // Moondancer ist der Torhueter
        	  if (see(flop))
        	  {
        		  c1 = warp(flop.X);
        		  c2 = warp(ball.X);
        		  c3 = Complex.sub(c2,c1).mul(Complex.expi(Math.PI/32));
        		  c4 = Complex.add(c1,Complex.mul(c3,DANGERRADIUS/c3.abs()));
        		  v1 = sphere.warp(c4);
        		  if ((distanceTo(ball)<3*distanceBallUnit) || (distance(ball,flop)<DANGERRADIUS)) { // Gefahr, weg mit dem Ball!
        			  clear(ball.X.c);
        			  actionMessage = "clear";
        		  }
        		  else if (sphere.diff(v1,self.X.c)>0.2*radius) { // stell Dich in die Verbindunglinie ball - flop
        			  goTo(v1,50);
        			  actionMessage = "position";
        		  }
        		  else { // dreh Dich und beobachte
        			  rotate();
        			  actionMessage = "rotate";
        		  }
        		  debug (new Circle(flop.X.c,DANGERRADIUS/sphere_rad));
        		  debug (new Line(new Vector(0,0,0),v1));
        	  }
        	  else { // dreh Dich und such den flop
        			  rotate();
        			  actionMessage = "search";
        	  }
          break;
          case MADDRAGON: // Mad Dragon &
          case ANGELFIRE: // Angelfire gehen zum Ball
            Puck partner = null;
            if (identity == MADDRAGON) partner = friend[ANGELFIRE]; else partner = friend[MADDRAGON];
            if (partner!=null) {
              if (distanceTo(ball)<distance(partner,ball)) { // ich bin dichter als der andere -> gehe zum Ball
                shoot(ball.X.c,goal.X.c,0.01);
                actionMessage = "shoot";
              }
              else { // gehe auf die gegenberliegende Seite und warte auf Torschuss
                c1 = warp(goal.X);
                c2 = warp(ball.X);
                c3 = Complex.sub(c1,c2);
                c4 = Complex.add(c1,Complex.mul(c3,DANGERRADIUS/c3.abs()));
                v1 = sphere.warp(c4);
                if (sphere.diff(v1,self.X.c)>4*radius) {
                  goTo(v1,50);
                  actionMessage = "lurk";
                }
                else {
                  rotate();
                  actionMessage = "rotate";
                }
                debug (new Line(new Vector(0,0,0),v1));
              }
            }
          break;
        }
      }
      action(P,L,R,""+identity); // abschicken der Aktionen
       // ----------------- Debugging ------------------------
      debug (new Text(null,name,1,0));
      debug (new Text(null,actionMessage,2,0));
      debug (new Text(null,"reference: "+reference,3,0));
      // ------------------- Remember -----------------------
     } // look
  } // loop

  /*************************************************************************
   * function clear: Vector
   ************************************************************************/

  private void clear(Vector v) {
    if (distanceTo(v)<1.5* distanceBallUnit)
      P = charge_max;
    else {
      P = 0;
      goTo(v,1);
    }
  }

 /*************************************************************************
  * function shoot: Vector, Vector, double
  ************************************************************************/

  private void shoot(Vector object, Vector target, double angle) {
  // Schiesst ein Object in eine gewisse Richtung target,
  // wenn der Wingel angle erreicht ist.
    Complex oc = sphere.warp(object);
    Complex tc = sphere.warp(target);
    if ((distanceTo(object)>2*distanceBallUnit))
    {
        P = 0;
        goTo(object,1);
    }
    else if ((Math.abs(oc.arg()-tc.arg())>angle))
    {
    	boolean opponentNear = false;
    	for (int i=0; i<nbOpponents; i++) if (distance(object,opponent[i].X.c)<1.5*distanceBallUnit) opponentNear = true;
    	if (opponentNear) P = charge_max; // clear is there is an opponent
    		else P = charge_min/100.0;
        turnTo(goal.X.c);
    }
    else
    {
    	P = charge_max;
    }
  }

  /**************************************************************************
   * function avoid: Vector
   *************************************************************************/

  private Vector avoid (Vector v) {
  // Man muss goal und flop vermeiden, wenn man auf v zufhrt
	Vector x = avoid(v,goal);
    if (see(flop)) x = avoid(x,flop);
    return x;
  }

  private Vector avoid(Vector v, Puck p) {
  // Vermeidet den Puck p
    if (p!=null) if (distanceTo(p)<AVOIDRADIUS) {
    // Hindernis umfahren
      Complex c1 = warp(p.X);
      Complex c2 = sphere.warp(v);
      double ang = triangle(new Complex(),c1,c2);
      if (Math.abs(ang)>Math.PI/2)
        return sphere.warp(Complex.add(Complex.sub(c2,c1).mul(Complex.expi(-ang/2)),c1));
    }
    return v;
  }

  /**************************************************************************
   * function goTo: Vector
   *************************************************************************/

  private void goTo(Vector v, double brake) {
  // der Spieler fhrt zum Vector v und vermeidet dabei goal und flop

    Vector v_ = avoid(v);
    double angle = angle(v_)/Math.PI; //angle in [-1,1]
    if (Math.abs(angle)>0.05) turnTo(v_); // Ziel nicht vor mir -> dreh dich
    else { // geradeaus volle Power
      double d = distanceTo(v_);
  	  L = engine_max*d*d/brake;
   	  R = engine_max*d*d/brake;
     }
  }

  /**************************************************************************
   * function turnTo: Vector
   *************************************************************************/

  private void turnTo(Vector v) {
    double angle = angle(v)/Math.PI; //angle in [-1,1]
    L = -100*engine_max*angle;
    R = -L;
  }

  /**************************************************************************
   * function rotate: no argument
   *************************************************************************/
  private void rotate() {
      L = engine_max;
      R = engine_min;
  }

  // ************************ Weitere Hilfsfunktionen *****************************

  private Complex warp (Matrix M) {
  // 3D-Matrix -> 2D komplexer Vektor
    Vector v = sphere.warp(M);
    return new Complex(v.x,v.y);
  }

  private double angle(Vector v) {
  // berechnet den Winkel zum Vector v
    if (v!=null) return new Complex(v).arg(); else return 0;
  }

  private double triangle(Complex a,Complex v,Complex b) {
  // berechnet den Winkel zwischen Vector v und a und b, v am Winkel
    if ((v!=null) && (a!=null) && (b!=null)) {
      Complex va = Complex.sub(a,v);
      Complex vb = Complex.sub(b,v);
      double x = vb.arg()-va.arg();
      if (x>Math.PI) x-=2*Math.PI;
      else if (x<-Math.PI) x+=2*Math.PI;
      return x;
    }
    else return 2*Math.PI;
  }

  private double distance(Vector v1, Vector v2) {
  // berechnet die Distanz zwischen den Vektoren v1 und v2
    if ((v1!=null) && (v2!=null)) return sphere.diff(v1,v2); else return sphere.rad*7;
  }

  private double distance(Puck p1,Puck p2) {
  // berechnet die Distanz zwischen den Pucks p1 und p2
    if ((p1!=null) && (p2!=null)) return distance(p1.X.c,p2.X.c); else return sphere.rad*7;
  }

  private double distanceTo(Puck p) {
  // berechnet die Distanz zwischen mir und Puck p
    if (p!=null) return distance(self.X.c,p.X.c);
    else return sphere.rad*7;
  }

  private double distanceTo(Vector v) {
  // berechnet die Distanz zwischen mir und Vector v
    return distance(self.X.c,v);
  }

  private boolean see(Puck p)
  {
	  if (p==null) return false;
	  if (p.X==null) return false;
	  return true;
  }

  private void detect(Puck[] pucks) {
    // Sucht die Pucks aus dem Array pucks zusammen und speichert alles schn ordentlich
    // in den Variablen ball, goal, flop, friend, opponent, node

    // Korrigiere die bekannten Daten anhand eines Knotens
	Matrix m = null;
	for (int i=0; i<6; ++i)
		if ((m==null) && (node_[i]!=null)) if (see(puck(NODE,0,i+1))) {
			reference = "node"+(i+1);
			m = Matrix.inv(node_[i].X).mul(puck(NODE,0,i+1).X);
			break;
		}

    if (m!=null) // muesste eigentlich immer der Fall sein, wenn unit.vision = 0.66
    {
        if (see(ball)) ball.X = ball.X.mul(m); /*immutable*/
        if (see(goal)) goal.X = goal.X.mul(m); /*immutable*/
        if (see(flop)) flop.X = flop.X.mul(m); /*immutable*/
        for (int k=0; k<3; ++k) if (see(friend[k])) friend[k].X = friend[k].X.mul(m); /*immutable*/
        for (int k=0; k<nbOpponents; ++k) if (see(opponent[k])) opponent[k].X = opponent[k].X.mul(m); /*immutable*/
        for (int k=0; k<6; ++k) if (see(node[k])) node[k].X = node[k].X.mul(m); /*immutable*/
    }

    // setze die alten Knoten auf null
    for (int i=0; i<6; i++) node_[i] = null;

    // ueberschreibe mit eventuellen neuen Daten
    for (int i=0; i<pucks.length; ++i) {
    	if (pucks[i].X != null) // ich muss sie auch sehen, nicht nur hoeren
    	{
    		if (pucks[i].equals(BALL))
    		{ // ein Ball ist's
    			if (pucks[i].t==0) ball = new Puck(pucks[i]);
    			else if (pucks[i].t==self.t) goal = new Puck(pucks[i]); // es ist ein Tor
    			else flop = new Puck(pucks[i]); // es ist ein Flop
    		}
    		else if (pucks[i].equals(UNIT))
    		{ // eine Unit ist's
      			if (pucks[i].t==self.t) {
      				if (pucks[i].message.length()>0)
      					friend[Integer.parseInt(pucks[i].message)] = new Puck(pucks[i]); // ... noch dazu aus meinem Team
      			}
      			else
      			{ // ... vom Gegnerteam
      				boolean found = false;
      				for (int j=0; j<nbOpponents; ++j) if (opponent[j].id.equals(pucks[i].id))
      				{
      					opponent[j] = new Puck(pucks[i]);
      					found = true;
      				}
      				if ((!found) && (nbOpponents<3)) opponent[nbOpponents++] = new Puck(pucks[i]);
      			}
    		}
    		else if (pucks[i].equals(NODE)) // ein Knoten ist's
    		{
    			node[pucks[i].n-1] = new Puck(pucks[i]);
    			node_[pucks[i].n-1] = new Puck(pucks[i]);
    		}
    	}
    }
    // vielleicht ist pucksLocated jetzt true?
    if (!pucksLocated) pucksLocated = locate();
}

  boolean locate() {
  // der Spieler kann erst dann loslegen, wenn er gewisse Pucks lokalisiert hat
    if (ball==null) return false;
    if (goal==null) return false;
    for (int i=0; i<3; i++) if (friend[i]==null) return false;
    for (int i=0; i<6; i++) if (node[i]==null) return false;
    return true;
  }
}
