package hoverball.team.LasMaquinaTeam_;

import hoverball.math.Complex;
import hoverball.math.Vector;


/**
 * Les actions d'un agent
 * @author Cyrille COIN, Florian COLIN, Thibaud COMTE
 *
 */
public class Action {

	/** 
	 * Référence vers un agent de notre équipe
	 */
	private Agent agent;


	/**
	 * Constructeur de classe
	 * @param maquina Référence vers le joueur de l'équipe LasMaquina
	 */
	public Action(Agent maquina) {
		this.agent = maquina;
	}

	public void turn(double force)
	{
		agent.L = agent.croyance.engine_max*force;
		agent.R = agent.croyance.engine_min*force;
	}

	public void turnTo(Vector v)
	{
		double angle = new Complex(v).arg()/Math.PI; 	//angle in [-1,1]
		agent.L = -100 * agent.croyance.engine_max*angle/2;
		agent.R = 0;	// Optimisation
	}
		
	/**
	 * Aller vers un endroit
	 * @param targetWarp
	 */
	public void moveToPos(Complex targetWarp)
	{
		Vector targetVector = agent.croyance.sphere.warp(targetWarp);
		if (agent.croyance.distanceTo(targetVector) > 5.0)
		{
			double angle = new Complex(targetVector).arg();
			if (Math.abs(angle)>0.20) turnTo(targetVector);	// On se repositionne en cas d'éloignement de la cible
			else { 
				double d = agent.croyance.distanceTo(targetVector);
				agent.L = agent.croyance.engine_max*d*d/500;
				agent.R = agent.croyance.engine_max*d*d/500;
			}
//			debug(new Line(self.X.c, targetVector));	// Trace une ligne entre le puck et la cible
		}
		else
		{
			agent.L = 0;
			agent.R = 0;
			agent.P = 0;
		}
	}


	/**
	 * Viser
	 * @param v
	 */
	public void viser(Vector v)
	{
		double angle = new Complex(v).arg()/Math.PI; //angle in [-1,1]
		agent.L = agent.croyance.engine_max/2 * angle;
		agent.R = 0;	// Optimisation
	}
	
	/**
	 * Viser et Tirer
	 * @param v
	 */
	public void shoot(Vector v)
	{
		double angle = new Complex(v).arg();
		if (Math.abs(angle) > 0.02) turnTo(v);	// On se repositionne en cas d'éloignement de la cible
		else { 
			double dist = agent.croyance.distanceTo(v);
			double maxDistanceShoot = 70.0;
			agent.P = agent.croyance.charge_max * (dist/maxDistanceShoot) * (dist/maxDistanceShoot);
			agent.state = State.HAVE_NO_BALL;
		}
	}

	/**
	 * Attraper la balle
	 */
	public void getBall() 
	{
		if (agent.croyance.distanceTo(agent.croyance.ball) > 2*agent.croyance.distanceBallUnit)
		{
			agent.P = 0;
			moveToPos(agent.croyance.sphere.warp(agent.croyance.ball.X.c));
		}
		else {
			agent.P = -0.01;
			agent.L = 0;
			agent.R = 0;
			agent.state = State.HAVE_BALL;
		}
	}
	
	
	/**
	 * Action: Faire le goal
	 * Le gardien se place entre le ballon et sa cage pour contrer les tirs
	 */
	public void doTheGoal()
	{
		// on récupère la position de la balle dans le plan 2D
		Complex ballWarp = agent.croyance.sphere.warp(agent.croyance.ball.X.c);
		// on récupère la position de notre cage dans le plan 2D
		Complex flopWarp = agent.croyance.sphere.warp(agent.croyance.flop.X.c);
		// on se repositionne sur l'origine du plan
		Complex targetWarp = Complex.sub(ballWarp, flopWarp);

		double distanceToGoal = 15.00;
		
		// Si la distance entre la cage et la balle est supérieur à distanceToGoal
		if (targetWarp.abs() > distanceToGoal)
		{
			// on normalise le complexe et on le multiplie par distanceToGoal
			targetWarp = targetWarp.norm().mul(distanceToGoal); /*immutable*/
			// On replace l'origine
			targetWarp = targetWarp.add(flopWarp); /*immutable*/

			// Mise à jour de l'endroit où aller
			agent.target = targetWarp;
		}
		else	// si on est près de la balle, on va vers la balle
		{
			agent.target = ballWarp;
			agent.P = 0.01;
		}
		moveToPos(agent.target);
	}
	

}
