package hoverball.team.LasMaquinaTeam_;

import hoverball.Puck;
import hoverball.math.Complex;
import hoverball.math.Sphere;
import hoverball.math.Vector;

/**
 * Les croyances de l'agent
 * @author Cyrille COIN, Florian COLIN, Thibaud COMTE
 *
 */
public class Croyance {

	/**
	 * Référence de l'agent
	 */
	private Agent agent;
	
	
	/**
	 * Information de base sur l'environnement
	 */
	
	/** La surface de la sphère */
	public Sphere sphere = null;
	/** sphere radius */
	public double sphere_rad = 0;
	/** min polarization */
	public double charge_min = 0;
	/** max polarization */
	public double charge_max = 0;
	/** min engine force */
	public double engine_min = 0;
	/** max engine force */
	public double engine_max = 0;
	/** vision angle 0..1 */
	public double vision = 0;
	/** distance Ball-Unit */
	public double distanceBallUnit = 0;
	/** Radius */
	public double radius = 0;
	
	
	/**
	 * Croyances sur l'ensemble des Pucks
	 */

	/** La balle */
	public Puck ball = null;
	/** Le but où marquer */
	public Puck goal = null;
	/** Le but des adversaires */
	public Puck flop = null;	
	/** Le nombre d'adversaire (3 max) */
	public int nbOpponents = 3;
	/** Le nombre de co-équipier (3 max) */
	public int nbFriends = 3;	
	/** Mon équipe, y compris moi */
	public Puck friend[] = new Puck[3];	
	/** L'équipe adverse */
	public Puck opponent[] = new Puck[3];
	/** Les noeuds de la sphère */
	public Puck node[] = new Puck[6];
	/** Les noeuds à la position vu la dernière fois */
	public Puck node_[] = new Puck[6];
	/** La position de l'agent à l'étape précédente */
	public Vector backupPosition = null;
	/** Le joueur démarre une action uniquement si il a localisé tous les pucks */
	public boolean pucksLocated = false;
	
	
	/**
	 * Variables rajoutées pour définir les désirs et rôles des agents de notre équipe
	 */
	
	/** Id du co-équipier le plus proche de la balle */
	public int friendBallClosest = -1;
	/** Id du co-équipier le plus proche de nos cages à défendre */
	public int friendFlopClosest = -1;
	/** Id du co-équipier qui n'est pas le plus proche de la balle et de la cage */ 
	public int friendGoalClosest = -1;
	/** Id de l'adversaire le plus proche de la balle */
	public int opponentBallClosest = -1;
	/** Id de l'adversaire le plus proche de nos cages à défendre */
	public int opponentFlopClosest = -1;
	/** Notre équipe a-t-elle la balle ? */
	public boolean friendHaveBall = false;
	/** L'équipe adverse a-t-elle la balle ? */
	public boolean opponentHaveBall = false;
	/** Le désir de chacun de nos équipiers */
	public Desir friendDesir[] = new Desir[3];
	/** L'intention de chacun de nos équipiers */
	public Intention friendIntention[] = new Intention[3];
	
	/** Distance de danger des cages */
	public double minDistanceToFlop = 8.00;
	/** L'agent est-il dangereusement près de ses cages ? */
	public boolean nearFlop = false;
	/** L'agent est-il dangereusement près des cages adverses ? */
	public boolean nearGoal = false;
	
	/** Un compteur pour savoir si l'agent fait du sur place ou pas */
	public int cpt = 1;
	/** La fréquence associée au compteur */
	public int frequency = 50;
	/** L'agent est-il statique */
	public boolean isStatic = false;

	/**
	 * Constructeur de classe
	 * @param agent référence vers l'agent
	 */
	public Croyance(Agent agent) {
		this.agent = agent;
	}

	
	/**
	 * Révision des croyances actuelles
	 */
	public void reviseBelief()
	{
		if (pucksLocated)
		{
			/**
			 * Trouver le co-équipier le plus proche de la balle
			 */
			double minDist = distance(friend[0], ball);
			friendBallClosest = 0;
			for (int i = 1; i < nbFriends; i++)	{
				if (distance(friend[i], ball) < minDist) {
					minDist = distance(friend[i], ball);
					this.friendBallClosest = i;
				}
			}

			/**
			 * Trouver le co-équipier le plus près de nos cages
			 * 
			 * Il faut que celui-ci ne soit pas le même que celui
			 * qui est près du ballon
			 */
			double[] distanceToFlop = new double[nbFriends];
			minDist = 100000;
			for (int i = 0; i < nbFriends; i++) 
			{
				if (i != friendBallClosest)
				{
					distanceToFlop[i] = distance(friend[i], flop);
					if (distanceToFlop[i] < minDist) {
						this.friendFlopClosest = i;
						minDist = distanceToFlop[i];
					}
				}
			}
			
			/**
			 * Attribution au dernier d'être le plus proche des cages adverses
			 */
			this.friendGoalClosest = 3 - this.friendBallClosest - this.friendFlopClosest;

			/**
			 * Trouver l'adversaire le plus proche de la balle
			 */
			minDist = distance(opponent[0], ball);
			this.opponentBallClosest = 0;
			for (int i = 1; i < nbOpponents; i++)	{
				if (distance(opponent[i], ball) < minDist) {
					minDist = distance(opponent[i], ball);
					this.opponentBallClosest = i;
				}
			}
			
			/**
			 * Trouver l'adversaire le plus près de nos cages
			 */
			minDist = distance(opponent[0], flop);
			this.opponentFlopClosest = 0;
			for (int i = 1; i < nbOpponents; i++)	{
				if (distance(opponent[i], flop) < minDist) {
					minDist = distance(opponent[i], flop);
					this.opponentFlopClosest = i;
				}
			}
			
			/**
			 * On met à jour les booléens permettant de savoir quelle
			 * équipe a la balle
			 */
			if (distance(friend[friendBallClosest], ball) < 2*distanceBallUnit) friendHaveBall = true;
			else friendHaveBall = false;
			
			if (distance(opponent[opponentBallClosest], ball) < 2*distanceBallUnit) opponentHaveBall = true;
			else opponentHaveBall = false;
			
//			/**
//			 * Met à jour le désir des co-équipiers
//			 */
//			for (int i = 0; i < nbFriends; i++) 
//			{
//				String[] token = friend[i].message.split(" ");
//				if (token.length > 0) {
//					int ordinal = Integer.parseInt( token[0] );
//					friendDesir[i] = Desir.values()[ordinal];
//				} else {
//					friendDesir[i] = Desir.NONE;
//				}					
//			}
			
			/**
			 * Si on est près de la balle, on se met dans l'état NEAR_BALL
			 */
			if (distanceTo(ball) < 2*distanceBallUnit) {
				if (agent.state != State.HAVE_BALL) agent.state = State.NEAR_BALL;
			} else {
				agent.state = State.HAVE_NO_BALL;
			}
			
			/**
			 * suis-je dans une zone dangereuse (proche de la cage) ?
			 */
			// Angle par rapport au flop
			double angleFlop = new Complex(flop.X.c).arg()/Math.PI;
			// angle par rapport au goal
			double angleGoal = new Complex(goal.X.c).arg()/Math.PI;
			
			// Si je suis proche des cages et que je vais vers celles-ci
			if (distanceTo(flop) < this.minDistanceToFlop && angleFlop < 0.05) nearFlop = true;
			else nearFlop = false;
			
			// Si je suis proche des cages et que je vais vers celles-ci
			if (distanceTo(goal) < this.minDistanceToFlop && angleGoal < 0.05) nearGoal = true;
			else nearGoal = false;
		}
		
		/**
		 * On incrémente le compteur si le puck n'a pas bougé par
		 * rapport à sa dernière position enregistrée
		 */
		if (backupPosition != null) {
			if (distanceTo(backupPosition) < 0.0001) {
				if (!isStatic) cpt = (cpt + 1) % frequency;
			}
			else cpt = 1;
		}
		// On met à jour la position sauvegardée.
		backupPosition = agent.getPosition();
		// Si le compteur est égal à 0, alors l'agent est statique
		if (cpt == 0) isStatic = true;
		else isStatic = false;
	}
	
	/**
	 * Est-ce que tous les pucks ont été détectés ?
	 * @author sb
	 * @return
	 */
	public boolean pucksLocated() {
		// the unit  only starts its task if all pucks are located
		if (ball == null) return false;
		if (flop == null) return false;
		if (goal == null) return false;
		for (int i = 0; i < nbFriends; i++) if (friend[i] == null) return false;
		for (int i = 0; i < nbOpponents; i++) if (opponent[i] == null) return false;
//		for (int i = 0; i < 6; i++) if (node[i] == null) return false;
		return true;
	}
	
	/** 
	 * Distance entre deux vecteurs 
	 */
	public double distance(Vector v1, Vector v2) {
		if ((v1 != null) && (v2 != null)) return sphere.diff(v1, v2); else return sphere.rad*7;
	}

	/**
	 * Retourne la distance, sur la sphère, entre deux pucks
	 * @param sphere
	 * @param p1
	 * @param p2
	 * @return
	 */
	public double distance(Puck p1, Puck p2) {
		if ((p1 != null) && (p2 != null)) return distance(p1.X.c, p2.X.c); else return sphere.rad*7;
	}
	
	/**
	 * Distance entre moi et le puck p
	 * @param p
	 * @return
	 */
	public double distanceTo(Puck p) {
		if (p != null) return distance(agent.getPosition(), p.X.c);
		else return sphere.rad*7;
	}

	/**
	 * Distance entre moi et un vecteur v
	 * @param v
	 * @return
	 */
	public double distanceTo(Vector v) {
		return distance(agent.getPosition(), v);
	}

	/**
	 * Returne un point à la perpendiculaire de la droite passant par les 2 points
	 * from et to. Ce point se trouve à la distance "distance" du point from
	 * @param sphere
	 * @param from
	 * @param to
	 * @param distance
	 * @return un point dans le plan complexe
	 */
	public Complex getPerpendicularPoint(Puck from, Puck to, double distance)
	{
		// on récupère la position de from
		Complex fromWarp = sphere.warp(from.X.c);
		// on récupère la position de to
		Complex toWarp = sphere.warp(to.X.c);
		// on se repositionne sur l'origine du plan
		Complex targetWarp = Complex.sub(fromWarp, toWarp);

		// on normalise le complexe et on le multiplie par e(iPi/2)
		targetWarp = targetWarp.norm().mul(new Complex(0,1)); /*immutable*/
		targetWarp = targetWarp.mul(distance); /*immutable*/
		// On replace l'origine
		targetWarp = targetWarp.add(toWarp); /*immutable*/

		return targetWarp;
	}
	
	/**
	 * Retourne vrai si on peut tirer, i.e. il n'y a pas d'obstacle dans la trajectoire
	 * sinon faux
	 * @return vrai si on peut tirer, i.e. il n'y a pas d'obstacle dans la trajectoire
	 * sinon faux
	 */
	public boolean canIShoot()
	{
		double maxAngleOpponent = 0.05;
		double distanceToGoal = distanceTo(goal);
		for (int i = 0; i < nbOpponents; i++)
		{
			double distanceToOpponent = distanceTo(opponent[i]);
			double angleOpponent = Math.abs(sphere.warp(opponent[i].X.c).arg());
			if (distanceToOpponent < distanceToGoal) {
				if (angleOpponent < maxAngleOpponent) {
					return false;
				}
			}
		}
		return true;
	}
	
	/**
	 * Retourne le message destiné à l'agent si il y en a un, sinon retourne null
	 * Si plusieurs messages lui sont destinés, on choisit le plus prioritaire.
	 * @return le message destiné à l'agent si il y en a un, sinon retourne null
	 */
	public Message readMsgs()
	{
		// On parcourt la liste des coéquipiers
		for (int i = 0; i < nbFriends; i++)
		{
			// on ne lie pas nos propre message...
			if (i != agent.identity) {
				// Si la chaine est non vide, on récupère le message
				if (friend[i] != null) {
					if (!friend[i].message.equalsIgnoreCase("") || friend[i].message != null) {
						Message msg = new Message(friend[i].message);
						// Si on est le destinataire, on retourne le message
						if (msg.getRecepteurId() == agent.identity) return msg;
					}
				}
			}
		}
		// Aucun message ne nous est destiné, on retourne null
		return null;
	}
	
}
