package hoverball.team.LasMaquinaTeam_;

import hoverball.math.Complex;

/**
 * Messages échangés entre les agents
 * Ils permettent de donner des ordres aux autres agents
 * @author Cyrille COIN, Florian COLIN, Thibaud COMTE
 *
 */
public class Message {

	/**
	 * L'id de l'agent émetteur du msg
	 */
	private int emmeteurId;
	
	/**
	 * L'id de l'agent recepteur du msg
	 */
	private int recepteurId;
	
	/**
	 * L'intention que l'on veut que l'agent recepteur ait 
	 */
	private Intention intention;
	
	/**
	 * La cible du désir
	 */
	private Complex target;
	
	
	/**
	 * Constructeur de classe
	 * @param idE
	 * @param idR
	 * @param d
	 * @param c
	 */
	public Message(int idE, int idR, Intention intention, Complex c) {
		this.emmeteurId = idE;
		this.recepteurId = idR;
		this.intention = intention;
		
		if (c == null) this.target = new Complex(0.000, 0.000);
		else this.target = c;
	}
	
	/**
	 * Construit un message a partir d'une chaîne de caractères
	 * @param msg
	 */
	public Message(String msg) 
	{
		// on divise la chaine en tokens
		String[] token = msg.split(" ");
		
		// si le message est vide, on construit un msg vide
		if (token.length != 5)
		{
			emmeteurId = -1;
			recepteurId = -1;
			intention = Intention.NONE;
			target = null;
		}
		else
		{
			emmeteurId = Integer.parseInt(token[0]);
			recepteurId = Integer.parseInt(token[1]);
			
			int ordinal = Integer.parseInt(token[2]);
			intention = Intention.values()[ordinal];
			
			double x = Double.parseDouble(token[3]);
			double y = Double.parseDouble(token[4]);
			target = new Complex(x, y);
		}		
	}
	
	/**
	 * Retourne le message sous forme de chaine de caractères
	 */
	public String toString() 
	{
		String msg = "";
		msg += Integer.toString(emmeteurId) + " ";
		msg += Integer.toString(recepteurId) + " ";
		msg += Integer.toString(intention.ordinal()) + " ";
		msg += Double.toString(Math.floor(target.x*1000.0) / 1000.0) + " ";
		msg += Double.toString(Math.floor(target.x*1000.0) / 1000.0);
		return msg;
	}

	public int getEmmeteurId() {
		return emmeteurId;
	}

	public void setEmmeteurId(int emmeteurId) {
		this.emmeteurId = emmeteurId;
	}

	public int getRecepteurId() {
		return recepteurId;
	}

	public void setRecepteurId(int recepteurId) {
		this.recepteurId = recepteurId;
	}

	public Intention getIntention() {
		return intention;
	}

	public void setIntention(Intention intention) {
		this.intention = intention;
	}

	public Complex getTarget() {
		return target;
	}

	public void setTarget(Complex target) {
		this.target = target;
	}
}
