/* Hoverball Competition 2007
 *
 * Team   : Team Total
 * Author : Martin Gloderer
 */

package hoverball.team.TeamTotal_;

import java.util.Map;
import java.util.Random;
import java.util.concurrent.ConcurrentHashMap;

import hoverball.*;
import hoverball.math.*;
import hoverball.debug.*;

public class TeamTotal extends Team
{
  // ---- Application

  public static void main (String[] args)
  {
    Application.setBounds();
    
    new TeamTotal("Team Total").show();
  }

  // ---- Constructor

  public TeamTotal() { this(null); }
  public TeamTotal(String team) { this(team,0xFF4400); }
  public TeamTotal(int color) { this(null,color); }
  public TeamTotal(String team, int color)
  {
    super((team == null || team == "")? "Team Total" : team,
       null,"Martin Gloderer",null);

    add(new MGL("Me",color));
    add(new MGL("Myself",color));
    add(new MGL("I",color));
  }
}

//---------------------------------------------------------------

class MGL extends Unit
{
	Sphere sphere;
	Behavior behavior;
	Behavior shooter;
	Behavior goalie;
	Behavior passer;
	Map<String, Puck> mypucks = new ConcurrentHashMap<String, Puck>();
	Map<Integer, Integer> shots = new ConcurrentHashMap<Integer, Integer>();

	Matrix pov = null;
	Matrix dest = null;

	String ballId;
	String goalId;
	String leaderId;

	double ballToGoal;
	double ballToEnemyGoal;
	double selfToBall;
	double leaderToBall;
	double enemyToBall;

	double Qmax;
	double Fmax;

	int team;
	double radius;
	double myEnergy;

	Matrix selfMatrix;
	Puck enemyGoal;
	Puck leader;

	boolean justShot;
	int sinceShot = 0;

	public MGL(String name, int color) {
		super("MGL", name, color);
		shooter = new Shooter(this);
		goalie = new Goalie(this);
		passer = new Passer(this);
		behavior = shooter;
	}

	public void loop()
	{
		sphere = new Sphere(option("world.radius"));  // Parameter auslesen...
		Qmax = option("unit.charge.max");
		Fmax = option("unit.engine.max");

		Random rand = new Random();
		Matrix from = null;
		Matrix to = new Matrix();

		ballToGoal = 2*Math.PI*sphere.rad;
		selfToBall = 2*Math.PI*sphere.rad;

//		behavior = new Clumsy(this);

		while (look())
		{
			if (self.n == 1 && behavior != shooter) behavior = shooter;
			if (self.n == 2 && behavior != goalie) behavior = goalie;
			if (self.n == 3 && behavior != passer) behavior = passer;

			pov = getPOV();
			selfMatrix = self.X;
			myEnergy = this.energy;

			if (justShot) {
//				System.out.println("I just shot!");
				sinceShot = 50;
				justShot = !justShot;
			}

			sinceShot--;

			team = self.t;
			radius = self.r;

			ballId = "ball 0 1";
			goalId = "ball " + self.t + " 1";
			leaderId = "unit " + self.t + " 1";

			updatePucks ();

			behavior.action();

//			double[] coeff = behavior.action();
//			action(coeff[0]*Qmax, coeff[1]*Fmax, coeff[2]*Fmax);

//			System.out.println(mypucks.get(ballId));

/*			for (int i=0; i<pucks.length; ++i) {
				Puck puck = pucks[i];
				if (puck.X == null) continue;
				debug(new Circle(puck.X.c,Math.PI/18));
				debug(new Text(puck.X.c,puck.id,4),-1);
			}
*/
			debug (new Text(self.X.c, (int) (this.energy*100) + "%", 2), -1);
			debug (new Text(self.X.c, new Integer(sinceShot).toString(), 4), -1);
			try {
				selfToBall = sphere.diff(self.X.c, getRel(mypucks.get(ballId).X).c);
				leaderToBall = sphere.diff(mypucks.get(leaderId).X.c, getRel(mypucks.get(ballId).X).c);
//				if (selfToBall > self.r*3) justShot = false;
//				debug (new Text (self.X.c, "THAT'S ME!", 9), -1);
				debug (new Circle (getRel(mypucks.get(ballId).X).c, Math.PI/30));
				debug (new Circle (getRel(mypucks.get(ballId).X).c, Math.PI/40));
				debug (new Text(pov.c, "POV", 2), -1);
				debug (new Text(getRel(mypucks.get(goalId).X).c, "goal", 2), -1);
//				debug (new Line (getRel(mypucks.get(ballId).X).c, getRel(mypucks.get(goalId).X).c), -1);
				ballToGoal = (int) sphere.diff(getRel(mypucks.get(ballId).X).c, getRel(mypucks.get(goalId).X).c);
				debug (new Text(getRel(mypucks.get(ballId).X).c, "dist: " + ballToGoal, 2), -1);
				debug (new Circle (dest.c, Math.PI/80));
				debug(new Line(self.X.c, dest.c), -1);
//				debug (new Text(self.X.c, mypucks.get(ballId).X.c.toString(), 9), -1);
				debug (new Line (getRel(mypucks.get(ballId).X).c, getRel(mypucks.get(goalId).X).c), -1);
				debug (new Text(getRel(mypucks.get(leaderId).X).c, "LEADER", -5), -1);
				debug (new Text(getRel(mypucks.get("unit 1 2").X).c, "GOALIE", -5), -1);
				debug (new Line (getRel(mypucks.get(ballId).X).c, getRel(mypucks.get(enemyGoal.id).X).c), -1);
			} catch (NullPointerException err) {
				debug (new Text(self.X.c, "NullPointerException!", 10), -1);
			};
		}
	}

	public void doAction(double q, double l, double r) {
		try {
			debug (new Text(self.X.c, (int) (q*100) + "  " + (int) (l*100) + "  " + (int) (r*100), 3), -1);
		} catch (NullPointerException err) {
			//System.out.println("caught error in action arguments.");
		};
		action (q*Qmax, l*Fmax, r*Fmax);
	}

	public double calcForce(int dist) {
		double force = Math.pow((double) dist /40, 2.0);
		if (force < 0.2) force = 0.2;
		return force;
	}

	// done
	public void updatePucks () {
		for (Puck puck : pucks) {
			if (puck.X == null) continue;				// Puck wird nur gehört
			mypucks.put(puck.id, new Puck(puck));
			mypucks.get(puck.id).X = getAbs(mypucks.get(puck.id).X);
			debug (new Circle(puck.X.c, Math.PI/50), -1);
//			if (puck.id.equals(ballId)) debug(new Text(self.X.c, "ball visible", 5), -1);
			if (puck.what == BALL && puck.t != self.t) enemyGoal = new Puck(puck);
			if (puck.what == UNIT && puck.t == self.t && puck.n == 1) leader = new Puck(puck);
		}
		int k = 2;
		for (String id : mypucks.keySet()) {
			Puck puck = mypucks.get(id);
//			debug (new Text(self.X.c, puck.id + " - " + puck.toString(), k), -1);
			debug (new Text(getRel(puck.X).c, id, 1), -1);
//			if (puck.id.equals(ballId)) debug(new Text(self.X.c, "ball seen at " + mypucks.get(ballId).X.c.toString(), 6), -1);
			k++;
		}
	}

	// done
	public Matrix getRel(Matrix m) {
		return new Matrix(pov).mulL(m);
	}

	// done
	public Matrix getAbs(Matrix m) {
		return Matrix.inv(pov).mulL(m);
	}

	// done
	public Matrix getPOV() {
		Matrix pov;
		Puck node = puck (NODE, 0, 1);
		if (node != null) {
			pov = node.X;
//			debug (new Text(self.X.c, "NODE 1", 7), -1);
		}
		else {
			node = puck (NODE, 0, 6);
if (node == null) return new Matrix(); // fix NullPointerException
			pov = Matrix.neg(node.X);
			pov = new Matrix(pov.a,pov.b.neg(),pov.c); /*immutable*/  // pov.b = pov.b.neg();
//			debug (new Text(self.X.c, "NODE 2", 7), -1);
		};
		return pov;
	}

	public void showDebug(String msg) {
		debug (new Text(self.X.c, msg, 5), -1);
	}

	public void showDebug(String msg, int line) {
		debug (new Text(self.X.c, msg, line), -1);
	}
}


abstract class State {

	MGL unit;
	Behavior behavior;

	public State(Behavior behavior) {
		this.behavior = behavior;
		this.unit = behavior.unit;
	}

	public boolean doAction() {
		return false;
	}

	public String toString() {
		return null;
	}
}


class FindBallState extends State {

	public FindBallState(Behavior behavior) {
		super(behavior);
	}

	public boolean doAction() {
		double q, l, r;
		q = 0;
		l = -1;
		r = 1;

//		unit.mypucks.remove(unit.ballId);

		unit.updatePucks();
		if (unit.mypucks.get(unit.ballId) != null && unit.mypucks.get(unit.goalId) != null) {
			behavior.state = behavior.stateGetBall;
			return false;
		}
		unit.doAction(q, l, r);					// sieht keinen Ball? drehe!
		return true;
	}

	public String toString() {
		return "FindBallState";
	}
}


class GetBallState extends State {

	public GetBallState (Behavior behavior) {
		super(behavior);
	}

	public boolean doAction () {
		double q, l, r;
		q = 0;
		l = -1;
		r = 1;

/*		// DEBUG!
		System.out.println("doing default action...");
		unit.doAction(0, 1, 1);
		return true;
*/
		if (unit.sinceShot <= 0 && unit.selfToBall < (unit.mypucks.get(unit.ballId).r + unit.radius)*2) {
//			System.out.println("changing state to HasBall...");
			behavior.state = behavior.stateHasBall;
			return false;
		}
		Vector a = Vector.vec(unit.selfMatrix.c, unit.getRel(unit.mypucks.get(unit.ballId).X).c);
//		a.add(Vector.vec(unit.selfMatrix.c, unit.getRel(unit.mypucks.get(unit.ballId).X).c));
//		Matrix X = Matrix.rot(a,(unit.radius + unit.mypucks.get(unit.ballId).r)/unit.sphere.rad).mulL(unit.getRel(unit.mypucks.get(unit.ballId).X));
		Matrix X = Matrix.rot(a, 0).mulL(unit.getRel(unit.mypucks.get(unit.ballId).X));
		Complex x = unit.sphere.warp(X.c);
		unit.dest = new Matrix(X);

//		unit.showDebug(new Integer((int) (x.abs())).toString(), 6);
		unit.showDebug(new Integer((int) (x.abs())).toString() + " - " + new Double(x.arg()).toString(), 6);
		l = -x.arg() + Math.max(0.2,1-10*Math.abs(x.arg())); // drehe und
		r =  x.arg() + Math.max(0.2,1-10*Math.abs(x.arg())); // fahre zu X
		q = 0;

		unit.doAction(q, l, r);
		return true;
	}

	public String toString() {
		return "GetBallState";
	}
}


class HasBallState extends State {

	public HasBallState (Behavior behavior) {
		super(behavior);
	}

	public boolean doAction () {
		double q, l, r;
		q = 0;
		l = -1;
		r = 1;

/*		if (unit.selfToBall < (unit.mypucks.get(unit.ballId).r + unit.radius)*2) {
			System.out.println("changing state to HasBall...");
			behavior.state = behavior.stateHasBall;
			return false;
		}
*/		Vector a = Vector.vec(unit.selfMatrix.c, unit.getRel(unit.mypucks.get(unit.goalId).X).c);
//		Matrix X = Matrix.rot(a,(unit.radius + unit.mypucks.get(unit.ballId).r)/unit.sphere.rad).mulL(unit.getRel(unit.mypucks.get(unit.ballId).X));
		Matrix X = Matrix.rot(a, 0).mulL(unit.getRel(unit.mypucks.get(unit.goalId).X));
		Complex x = unit.sphere.warp(X.c);
		unit.dest = new Matrix(X);

		unit.showDebug(new Integer((int) (x.abs())).toString() + " - " + new Double(x.arg()).toString(), 6);
		if (unit.selfToBall > unit.radius*5) {
			behavior.state = behavior.stateGetBall;
		}
		if (unit.myEnergy > 0.2) {
			q = -0.01 - ((int) unit.myEnergy / 10)*0.02;
//			q = 0.05;
		} else q = 0.5;
		if (Math.abs(x.arg()) < 0.02/Math.sqrt(x.abs())) {
			q = unit.calcForce((int) (x.abs()));
			behavior.state = behavior.stateGetBall;
			unit.justShot = true;
		}
		if (unit.selfToBall < (unit.radius + unit.mypucks.get(unit.ballId).r)*2) {
			l = -x.arg();
			r = x.arg();
		} else {
//			l = -x.arg()*5;
//			r = x.arg()*5;
			l = 0;
			r = 0;
		}
//		l = -x.arg() + Math.max(0.2, 1-10*Math.abs(x.arg()));
//		r =  x.arg() + Math.max(0.2, 1-10*Math.abs(x.arg()));

//		System.out.println("I am now in HasBall state.");
		unit.doAction(q, l, r);
//		unit.mypucks.remove(unit.ballId);		// find ball again
		return true;
	}

	public String toString() {
		return "HasBallStatePasser";
	}
}


class Shooter extends Behavior {

//	MGL unit;
	boolean todo;

	public Shooter (MGL unit) {
		super(unit);
		try {
			unit.mypucks.clear();
			unit.updatePucks();
		} catch (NullPointerException err) {};
		this.stateFindBall = new FindBallState(this);
		this.stateGetBall = new GetBallState(this);
		this.stateHasBall = new HasBallState(this);
		this.state = stateFindBall;
		this.todo = false;
	}

	public void setState(State state) {
		this.state = state;
	}

	public void action () {
		unit.showDebug("SHOOTER", 7);

		if (unit.mypucks.get(unit.ballId) == null || unit.mypucks.get(unit.goalId) == null) {
			state = stateFindBall;
		}

//		System.out.println("doing action in state " + state.toString());
//		while(todo) {
//			todo = !state.doAction();
//		}
		state.doAction();
//		System.out.println(state.toString());
		unit.showDebug(state.toString());

/*			if (unit.sinceShot <= 0 && unit.myEnergy > 30 && unit.sphere.diff(unit.selfMatrix.c, unit.dest.c) < unit.radius*5) {
				q = -0.5;
				l = l/4;
				r = r/4;
			}
			if (unit.sphere.diff(unit.selfMatrix.c, unit.dest.c) < unit.mypucks.get(unit.ballId).r/2) {
				q = unit.calcForce(unit.ballToGoal);
				unit.justShot = true;
			}
*/
/*				if (q > 0) {
				try {
					from = new Matrix(to);
					to = new Matrix(mypucks.get(ballId).X);
					dist = (int) (sphere.diff(from.c, to.c));
					if (dist == 0) continue;
					shots.put(new Integer((int) (q*100)), new Integer((int) dist));
					System.out.println(shots.size() + " shots; force: " + new Integer((int) (q*100)).toString() + ", dist: " + new Integer(dist).toString());
				} catch (NullPointerException err) {
					to = new Matrix(getRel(mypucks.get(ballId).X));
				}
			}

			for (int force : shots.keySet()) {
				debug (new Text(self.X.c, force + " => " + shots.get(force), (double) force /10 + 5), -1);
			}
			try {
				debug (new Line (getRel(from).c, getRel(to).c), -1);
//				System.out.println("line from " + from.c + " to " + to.c);
			} catch (NullPointerException err) {};

*///				action(q*Qmax,l*Fmax,r*Fmax);
//			action(q*Qmax, 0.7*Fmax, Fmax);
	}
}


class FindBallStatePasser extends State {

	public FindBallStatePasser (Behavior behavior) {
		super(behavior);
	}

	public boolean doAction() {
		double q, l, r;
		q = 0;
		l = -1;
		r = 1;

//		unit.mypucks.remove(unit.ballId);

		unit.updatePucks();
		if (unit.mypucks.get(unit.ballId) != null && unit.mypucks.get(unit.leaderId) != null) {
			behavior.state = behavior.stateGetBallPasser;
			return false;
		}
		unit.doAction(q, l, r);					// sieht keinen Ball? drehe!
		return true;
	}

	public String toString() {
		return "FindBallState";
	}
}


class GetBallStatePasser extends State {

	public GetBallStatePasser (Behavior behavior) {
		super(behavior);
	}

	public boolean doAction () {
		double q, l, r;
		q = 0;
		l = -1;
		r = 1;

/*		// DEBUG!
		System.out.println("doing default action...");
		unit.doAction(0, 1, 1);
		return true;
*/
		if (unit.sinceShot <= 0 && unit.selfToBall < (unit.mypucks.get(unit.ballId).r + unit.radius)*2) {
//			System.out.println("changing state to HasBall...");
			behavior.state = behavior.stateHasBallPasser;
			return false;
		}

		Vector a = Vector.vec(unit.selfMatrix.c, unit.getRel(unit.mypucks.get(unit.ballId).X).c);
//		a.add(Vector.vec(unit.selfMatrix.c, unit.getRel(unit.mypucks.get(unit.ballId).X).c));
//		Matrix X = Matrix.rot(a,(unit.radius + unit.mypucks.get(unit.ballId).r)/unit.sphere.rad).mulL(unit.getRel(unit.mypucks.get(unit.ballId).X));
		Matrix X = Matrix.rot(a, 0).mulL(unit.getRel(unit.mypucks.get(unit.ballId).X));
		Complex x = unit.sphere.warp(X.c);
		unit.dest = new Matrix(X);

//		unit.showDebug(new Integer((int) (x.abs())).toString(), 6);
		unit.showDebug(new Integer((int) (x.abs())).toString() + " - " + new Double(x.arg()).toString(), 6);
		l = -x.arg() + Math.max(0.2,1-10*Math.abs(x.arg())); // drehe und
		r =  x.arg() + Math.max(0.2,1-10*Math.abs(x.arg())); // fahre zu X
		q = 0;

		unit.doAction(q, l, r);
		return true;
	}

	public String toString() {
		return "GetBallStatePasser";
	}
}


class HasBallStatePasser extends State {

	public HasBallStatePasser (Behavior behavior) {
		super(behavior);
	}

	public boolean doAction () {
		double q, l, r;
		q = 0;
		l = -1;
		r = 1;

/*		if (unit.selfToBall < (unit.mypucks.get(unit.ballId).r + unit.radius)*2) {
			System.out.println("changing state to HasBall...");
			behavior.state = behavior.stateHasBall;
			return false;
		}
*/
		String targetId;
		if (unit.sphere.diff(unit.mypucks.get(unit.leaderId).X.c, unit.getAbs(unit.selfMatrix).c) < unit.ballToGoal) {
			targetId = unit.leaderId;
		} else targetId = unit.goalId;

		Vector a = Vector.vec(unit.selfMatrix.c, unit.getRel(unit.mypucks.get(targetId).X).c);
//		Matrix X = Matrix.rot(a,(unit.radius + unit.mypucks.get(unit.ballId).r)/unit.sphere.rad).mulL(unit.getRel(unit.mypucks.get(unit.ballId).X));
		Matrix X = Matrix.rot(a, 0).mulL(unit.getRel(unit.mypucks.get(targetId).X));
		Complex x = unit.sphere.warp(X.c);
		unit.dest = new Matrix(X);

		unit.showDebug(new Integer((int) (x.abs())).toString() + " - " + new Double(x.arg()).toString(), 6);
		if (unit.selfToBall > unit.radius*5) {
			behavior.state = behavior.stateGetBallPasser;
		}
		if (unit.myEnergy > 0.2) {
			q = -0.01 - ((int) unit.myEnergy / 10)*0.01;
//			q = 0.05;
		} else q = 0.5;
		if (Math.abs(x.arg()) < 0.1/Math.sqrt(x.abs())) {
			q = unit.calcForce((int) (x.abs()))/6;
			behavior.state = behavior.stateGetBallPasser;
			unit.justShot = true;
		}
		if (unit.selfToBall < (unit.radius + unit.mypucks.get(unit.ballId).r)*2) {
			l = -x.arg();
			r = x.arg();
		} else {
//			l = -x.arg()*5;
//			r = x.arg()*5;
			l = 0;
			r = 0;
		}
//		l = -x.arg() + Math.max(0.2, 1-10*Math.abs(x.arg()));
//		r =  x.arg() + Math.max(0.2, 1-10*Math.abs(x.arg()));

//		System.out.println("I am now in HasBall state.");
		unit.doAction(q, l, r);
//		unit.mypucks.remove(unit.ballId);		// find ball again
		return true;
	}

	public String toString() {
		return "HasBallStatePasser";
	}
}


class Passer extends Behavior {

//	MGL unit;
	boolean todo;

	public Passer (MGL unit) {
		super(unit);
		try {
			unit.mypucks.clear();
			unit.updatePucks();
		} catch (NullPointerException err) {};
		this.stateFindBallPasser = new FindBallStatePasser(this);
		this.stateGetBallPasser = new GetBallStatePasser(this);
		this.stateHasBallPasser = new HasBallStatePasser(this);
		this.state = stateFindBallPasser;
		this.todo = false;
		unit.doAction(0, -1, 1);
	}

	public void setState(State state) {
		this.state = state;
	}

	public void action () {
		unit.showDebug("PASSER", 7);

		if (unit.mypucks.get(unit.ballId) == null || unit.mypucks.get(unit.leaderId) == null) {
			state = stateFindBallPasser;
			return;
		}

		if (unit.sphere.diff(unit.mypucks.get(unit.leaderId).X.c, unit.getAbs(unit.selfMatrix).c) < unit.radius*10) {
			unit.showDebug("PANIC!", 8);
			unit.doAction(0, -1, 1);
			return;
		}

//		System.out.println("doing action in state " + state.toString());
//		while(todo) {
//			todo = !state.doAction();
//		}
		state.doAction();
//		System.out.println(state.toString());
		unit.showDebug(state.toString());

/*			if (unit.sinceShot <= 0 && unit.myEnergy > 30 && unit.sphere.diff(unit.selfMatrix.c, unit.dest.c) < unit.radius*5) {
				q = -0.5;
				l = l/4;
				r = r/4;
			}
			if (unit.sphere.diff(unit.selfMatrix.c, unit.dest.c) < unit.mypucks.get(unit.ballId).r/2) {
				q = unit.calcForce(unit.ballToGoal);
				unit.justShot = true;
			}
*/
/*				if (q > 0) {
				try {
					from = new Matrix(to);
					to = new Matrix(mypucks.get(ballId).X);
					dist = (int) (sphere.diff(from.c, to.c));
					if (dist == 0) continue;
					shots.put(new Integer((int) (q*100)), new Integer((int) dist));
					System.out.println(shots.size() + " shots; force: " + new Integer((int) (q*100)).toString() + ", dist: " + new Integer(dist).toString());
				} catch (NullPointerException err) {
					to = new Matrix(getRel(mypucks.get(ballId).X));
				}
			}

			for (int force : shots.keySet()) {
				debug (new Text(self.X.c, force + " => " + shots.get(force), (double) force /10 + 5), -1);
			}
			try {
				debug (new Line (getRel(from).c, getRel(to).c), -1);
//				System.out.println("line from " + from.c + " to " + to.c);
			} catch (NullPointerException err) {};

*///				action(q*Qmax,l*Fmax,r*Fmax);
//			action(q*Qmax, 0.7*Fmax, Fmax);
	}
}


class Goalie extends Behavior {

	public Goalie (MGL unit) {
		super(unit);
//		this.unit = unit;
	}

	public void action() {
		unit.showDebug("GOALIE", 7);

		if (unit.mypucks.get(unit.ballId) == null || unit.mypucks.get(unit.enemyGoal.id) == null || unit.mypucks.get(unit.leaderId) == null) {
			unit.doAction(0, -1, 1);
			return;
		}

		if (unit.sphere.diff(unit.mypucks.get(unit.leaderId).X.c, unit.getAbs(unit.selfMatrix).c) < unit.radius*10) {
			unit.showDebug("PANIC!", 8);
			unit.doAction(0, -1, 1);
			return;
		}

		double q, l, r;
		q = 0;
		l = -1;
		r = 1;

		Vector a = Vector.vec(unit.getRel(unit.mypucks.get(unit.enemyGoal.id).X).c, unit.getRel(unit.mypucks.get(unit.ballId).X).c);
//		Matrix X = Matrix.rot(a,(unit.radius + unit.mypucks.get(unit.ballId).r)/unit.sphere.rad).mulL(unit.getRel(unit.mypucks.get(unit.ballId).X));
		Matrix X = Matrix.rot(a, -a.abs()/2).mulL(unit.getRel(unit.mypucks.get(unit.ballId).X));
		Complex x = unit.sphere.warp(X.c);
		unit.dest = new Matrix(X);

//		unit.showDebug(new Integer((int) (x.abs())).toString(), 6);
		unit.showDebug(new Integer((int) (x.abs())).toString() + " - " + new Double(x.arg()).toString(), 6);
		l = -x.arg() + Math.max(0.2,1-10*Math.abs(x.arg())); // drehe und
		r =  x.arg() + Math.max(0.2,1-10*Math.abs(x.arg())); // fahre zu X
		l = unit.selfToBall*l/30;
		r = unit.selfToBall*r/30;

		unit.doAction(q, l, r);
	}
}


abstract class Behavior {

	MGL unit;
	State state;
	State stateFindBall;
	State stateGetBall;
	State stateHasBall;
	State stateFindBallPasser;
	State stateGetBallPasser;
	State stateHasBallPasser;

	public Behavior(MGL unit) {
		this.unit = unit;
	}

	public void action() {}
}
