/**
 * Contient l'intelligence de l'attaquant
 */
package hoverball.team.TribalQueens_;

import hoverball.math.Vector;

/**
 * Etre attaquant est un role de joueur.
 */
public class Attaquant extends Role {

	/**
	 * La distance maximale à laquelle on tir
	 */
	private double distanceTirMax = 100;

	/**
	 * La distance à partir de laquelle on arrete de tirer et on fonce sur le
	 * but
	 */
	// private double distanceTirMin = 0;

	/**
	 * L'angle de direction vers les buts
	 */
	private double angleTir = Math.PI / 12;

	/**
	 * L'angle minimal de précision pour faire une passe
	 */
	private double anglePasse = Math.PI / 24;

	/**
	 * vrai si on est reposé
	 */
	private boolean repos = true;

	/**
	 * L'energie en dessous de laquelle on ne doit pas descendre
	 */
	private double energieMin = 0.2;

	/**
	 * L'energie au dessus de laquelle on est en forme
	 */
	private double energieOK = 0.7;

	/**
	 * Le constructeur de l'attaquant.
	 * 
	 * @param joueur
	 *            Le joueur qui est l'attaquant.
	 * @param terrain
	 *            Le terrain sur lequel on joue.
	 */
	public Attaquant(TribalQueen joueur, Environnement terrain) {
		super(joueur, terrain);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see tribalQueens.Role#reflechir()
	 */
	public void reflechir() {
		critique = false;
		
		// on recupère les infos sur la balle
		double distanceBalle = distance(joueur.getBalle());
		double angleBalle = angle(joueur.getBalle());

		// on recupère les infos sur le but
		double distanceBut = distance(joueur.getButEquipe());
		double angleBut = angle(joueur.getButEquipe());

		// si on est reposé
		if (repose()) {

			// si on ne possède pas la balle
			if (!possedeBalle()) {

				// on va sur la balle en evitant les joueurs si la balle est
				// loin
				if (distanceBalle > distanceAttraction * 2) {
					allerA(joueur.getBalle(), false, angleMouvement, true);
				} else {
					// en bourrinant sinon (on se dispute la balle par exemple)
					allerA(joueur.getBalle(), false);
				}

				// si on peut attirer la balle
				if (attirableBalle()) {

					// si je suis trop pres du but adverse, je degage
					if (distance(joueur.getButAdverse()) < 10) {
						envoyer(100);
					} else {
						// sinon on essaie de l'attraper
						attraper(distanceBalle);
					}

				}

			} else {

				// si on arrive ici alors la balle est proche de nous, on la
				// possède
				// on conserve la balle
				attraper(distanceBalle);

				// on va au but, en force si on est poche du but
				if (distanceBut < distanceTirMax / 4) {
					allerA(joueur.getButEquipe(), true, Math.min(angleTir,
							angleTir * (distanceBut / distanceTirMax)), false);
				} else {
					// en douceur sinon (on drible les joueurs)
					allerA(joueur.getButEquipe(), false, Math.min(angleTir,
							angleTir * (distanceBut / distanceTirMax)), true);
				}

				// si un equipier est plus proche du but
				if (equipierPlusProche(joueur.getButEquipe())) {
					// alors on fait la passe
					passer();
				} else {
					// sinon on tir
					tirPrecis(angleBalle, angleBut, distanceBut);
				}

				// si je suis en danger et loin du but
				if ((adversaireTropProche() && distanceBut > distanceTirMax / 4)) {
					// j'essaie de passer
					if (peutPasser()) {
						passer();
					} else if (distanceBut > distanceTirMax / 2
							&& Math.abs(distance(joueur.getButAdverse())) > Math.PI/2) {
						// sinon je dégage si on est loin du but mais orienté
						// vers le but
						envoyer(100);
					}
				}

				// si je suis trop pres du but adverse, je degage s'il n'y a pas de dangers pour le but
				if (distance(joueur.getButAdverse()) < 20 && Math.abs(angle(joueur.getButAdverse()) - angleBalle) > 2 * angleTirPrecis(distance(joueur.getButAdverse())) ) {
					envoyer(100);
				}

			}
		}
	}

	/**
	 * Permet de savoir si la balle est suffisament proche pour etre attirée
	 * 
	 * @return Vrai si la balle est attirable, faux sinon
	 */
	private boolean attirableBalle() {
		// la balle est attirable si elle est dans notre champ d'action,
		// d'attraction
		return (distance(joueur.getBalle()) < distanceAttraction && Math
				.abs(angle(joueur.getBalle())) < angleAttraction);
	}

	/**
	 * Calcul de l'angle avec lequel on tir précisement
	 * 
	 * @param distanceBut
	 *            La distance qui nous sépare du but
	 * @return L'angle de précision minimum
	 */
	private double angleTirPrecis(double distanceBut) {
		/*
		 * On touche la balle si, en partant du principe que la balle va tout
		 * droit, elle forle la balle. Pour calculer l'angle de tir, on se
		 * repère à trois point :
		 * 
		 * - A le centre du joueur,
		 * 
		 * - B le centre du but visé,
		 * 
		 * - C le centre d'une balle imaginaire qui touche le but visé
		 * 
		 * Le triangle ABC est alors reectangle en B et nous cherchons l'angle
		 * A. Pour cela nous utilisons l'arctangeante avec
		 * 
		 * - AB = distance au but,
		 * 
		 * - BC = 2 * le rayon d'une balle
		 * 
		 * Enfin, du fait de l'utilisation de arctangeante, quand on est proche
		 * du but, une petite erreur de calcule implique une grande erreur dans
		 * l'angle, c'est pourquoi on reduit l'angle de 20%. Une meilleur
		 * solution est à trouver.
		 */
		return Math.atan(2.0 * terrain.getRayonBalle() / distanceBut) * 0.8;
	}

	/**
	 * Tir si on peut et qu'on est asser précis
	 * 
	 * @param angleBalle
	 *            L'angle vers la balle
	 * @param angleBut
	 *            L'angle vers le but
	 * @param distanceBut
	 *            La distance qui nous sépare du but
	 */
	private void tirPrecis(double angleBalle, double angleBut,
			double distanceBut) {

		// si on peut tirer ou si je ne peux pas faire de passe
		// ou si je suis proche du but
		if (!adversaireEntre(joueur.getButEquipe()) || !peutPasser()
				|| distanceBut < distanceTirMax / 2) {

			// si on est assez proche du but
			if (Math.abs(angleBalle) < anglePasse
					&& distanceBut < distanceTirMax
					// et qu'on a atteint la précision suffisante entre la balle
					// et le
					// but (possibilité de tir aveugles)
					&& Math.abs(angleBut - angleBalle) < Math
							.abs(angleTirPrecis(distanceBut))) {
				// on tir FORT
				envoyer(distanceBut * 2);
			}
		} else {
			// si je ne peux pas tirer je fait une passe
			passer();
		}
	}

	/**
	 * Permet de savoir si on est reposé
	 * 
	 * @return Vrai si on est reposé et faux sinon
	 */
	private boolean repose() {
		// si on a assez d'energie
		if (joueur.getEnergy() > energieOK) {
			// alors on est reposé
			repos = true;
		} else if (!repos || joueur.getEnergy() < energieMin) {
			// sinon si on est en train de recharger ou si on a un manque
			// d'energie alors on ne doit pas polariser la balle
			polarisation = 0;

			// et on bouge plus
			moteurDroit = 0;
			moteurGauche = 0;

			// on est pas reposé
			repos = false;
		}

		// on retourne la variable de repos
		return repos;
	}

	/**
	 * Fait une passe au joueur le plus proche de but qui soit démarqué
	 */
	private void passer() {
		// on recupère le joueur à qui on doit faire une passe
		Vector equipier = getSoutien();

		// on fait une passe entre lui et le but
		equipier = equipier.add(joueur.getButEquipe().sub(equipier).norm().mul(0.3)); /*immutable*/
		
		joueur.tag(equipier, "passe");
		
		// je me tourne vers lui
		tourner(angle(equipier));

		// si je suis assez precis pour que la balle tombe dans un rayon de 5
		// autour de mon copain
		if (Math.abs(angle(equipier) - angle(joueur.getBalle())) < Math
				.atan(5 / distance(equipier))) {
			// System.out.println(joueur.title+" passe");
			envoyer(distance(equipier));
		}

	}

	/**
	 * @return Le joueur le plus proche du but à qui on peut faire une passe
	 */
	private Vector getSoutien() {
		Vector[] equipiers = joueur.getEquipiers();
		Vector resultat = null;
		// on parcours tous les equipiers
		for (int i = 0; i < equipiers.length; i++) {
			// si c'est pas moi
			if (!equipiers[i].equals(joueur.getJoueur())
			// qu'il est pas désactivé
					&& !joueur.getEquipiersDesactive()[i]
					// qu'il est pas trop près
					&& distance(equipiers[i]) > 40
					// mais pas trop loin non plus
					&& distance(equipiers[i]) < distanceTirMax
					// ni trop près du but adverse
					&& terrain.distance(equipiers[i], joueur.getButAdverse()) > 20
					&& terrain.distance(equipiers[i], joueur.getButEquipe()) < distance(joueur.getButEquipe()) *0.75
					// qu'il n'y a pas de risque d'interseption
					&& !adversaireEntre(equipiers[i])
					// et qu'il est démarqué
					&& !adversaireProche(equipiers[i])) {
				// si on a personne à qui faire la passe
				if (resultat == null) {
					// on a une soutient
					resultat = equipiers[i];
				} else if (terrain.distance(resultat, joueur.getButEquipe()) > terrain
						.distance(equipiers[i], joueur.getButEquipe())) {
					// sinon on peut lui passer s'il est mieux placé que le
					// précédent
					resultat = equipiers[i];
				}
			}
		}
		// on retourne le joueur trouvé
		return resultat;
	}

	/**
	 * @return Vrai si on peut passer
	 */
	private boolean peutPasser() {
		// vrai si on a quelqu'un à qui passer
		return getSoutien() != null;
	}

	/**
	 * @param point
	 * @return Vrai s'il y a possibilité de passer à un joueur bien placé
	 */
	private boolean equipierPlusProche(Vector point) {
		// on recupère celui à quie on doit faire la passe
		Vector equipier = getSoutien();

		// s'il est plus pres du but d'au moins la moitié
		if (equipier != null
				&& terrain.distance(joueur.getButEquipe(), equipier)
						+ distance(joueur.getButEquipe()) / 4 < distance(joueur
						.getButEquipe())) {
			// alors on doit passer
			return true;
		}

		// sinon pas de passe
		return false;
	}

	/**
	 * @param point
	 * @return Vrai s'il y a un adversaire entre le joueur et le point
	 */
	private boolean adversaireEntre(Vector point) {
		// pour l'instant le passage n'est pas bloqué
		boolean bloque = false;
		Vector[] adversaires = joueur.getAdversaires();

		// pour tous les adversaires
		for (int i = 0; i < adversaires.length; i++) {
			// il bloque le chemin si son angle est proche de celui du point
			// visé et qu'il est moins loin
			bloque = bloque
					|| (Math.abs(angle(point) - angle(adversaires[i])) < anglePasse && distance(adversaires[i]) < distance(point));
		}

		return bloque;
	}

	/**
	 * @param point
	 * @return Vrai s'il y a un adversaire proche
	 */
	private boolean adversaireProche(Vector point) {
		// pour l'instant le passage n'est pas bloqué
		boolean bloque = false;
		Vector[] adversaires = joueur.getAdversaires();

		// pour tous les adversaires
		for (int i = 0; i < adversaires.length; i++) {
			// il bloque le chemin si son angle est proche de celui du point
			// visé et qu'il est moins loin
			bloque = bloque || terrain.distance(adversaires[i], point) < 15;
		}

		return bloque;
	}

	/**
	 * @return vrai si un adversaire et trop proche
	 */
	private boolean adversaireTropProche() {
		Vector[] adversaires = joueur.getAdversaires();
		// pour tous les adversaires
		for (int i = 0; i < adversaires.length; i++) {
			// s'il est proche et devant
			if (distance(adversaires[i]) < distancePossession * 2
					&& Math.abs(angle(adversaires[i])) < anglePossession) {
				// alors il est dangereux
				return true;
			}
		}
		// si on arrive ici c'est qu'il n'y a pas de risque
		return false;
	}

}
