/**
 * 
 */
package hoverball.team.TribalQueens_;

import hoverball.Puck;
import hoverball.math.Matrix;
import hoverball.math.Sphere;
import hoverball.math.Vector;

import java.io.PrintWriter;
import java.io.StringWriter;

/**
 * @author detantxavi
 * 
 */
public class Communication {

	/**
	 * Transforme une liste de pucks en une chaine de caractères
	 * 
	 * @param pucks
	 *            La liste des pucks à transformer
	 * @return Le liste en chaine de caractères < 100 caractères
	 */
	public static String pucksToString(Puck[] pucks) {
		String str = "";

		// pour tous les pucks
		for (Puck p : pucks) {
			// on les transforme en chaine et on les sépare avec un espace
			str += puckToString(p);
		}

		// on retourne la chaine
		return str;
	}

	/**
	 * Transforme un puck en une chaine d'exactement 20 caractères
	 * 
	 * @param p
	 *            Le puck à transformer
	 * @return le puck transformé
	 */
	private static String puckToString(Puck p) {
		String str = "";
		// on recupère le vecteur representant sa matrice
		StringWriter sw = new StringWriter();
		PrintWriter pw = new PrintWriter(sw);

		if (p != null) {

			// on ecris l'id du puck
			str += p.what;
			// si il est desactivé
			if (p.message != null && p.message.equals("desactive")) {
				str += 9;
			} else {
				str += p.t;
			}
			str += p.n;
			// si on a une matrice
			if (p.X != null) {
				// on la treansforme pour avoir une taille de 20
				Vector v = Sphere.euler(p.X);
				pw.printf("%+1.2f %+1.2f %+1.2f", v.x, v.y, v.z);
			} else {
				// si on a pas de matrice on envoie null
				pw.printf("%17s", "null");
			}

		} else {
			// si on arrive ici c'est que le puck est null
			pw.printf("%20s", "null");
		}

		// on ajoute la matrice
		str += sw.toString();

		// on retourne la chaine
		return str;
	}

	/**
	 * Transforme une chaine en un puck
	 * 
	 * @param s
	 *            La chaine decrivant le puck
	 * @return Le puck de la chaine
	 */
	private static Puck puckFromString(String s) {
		int what, t, n;
		Vector v;
		String[] vector;
		Matrix x;
		Puck res;
		if (!s.substring(0, 1).equals(" ")) {
			// on recupère l'id du puck
			what = Integer.parseInt(s.substring(0, 1));
			t = Integer.parseInt(s.substring(1, 2));
			n = Integer.parseInt(s.substring(2, 3));

			// si on nous a transmis la matrice
			if (!s.substring(3).contains("null")) {
				// on separe les nombres
				vector = s.substring(3).replace(",", ".").split(" ");
				// on creer le vecteur
				v = new Vector(Double.parseDouble(vector[0]), Double
						.parseDouble(vector[1]), Double.parseDouble(vector[2]));
				// on fait l'euler inverse
				x = Sphere.euler(v);
			} else {
				// si on a pas de matrice alors elle est nulle
				x = null;
			}
			res = new Puck(what, t, n, x, 0, 0, "");
		} else {
			// si on arrive ici c'est que le puck est null
			res = null;
		}
		// on crée le puck
		return res;
	}

	/**
	 * Transforme une chaine en pucks
	 * 
	 * @param s
	 *            La chaine contenant les pucks
	 * @return La liste des pucks contenus dans la chaine
	 */
	public static Puck[] pucksFromString(String s) {
		// le nombre de pucks à recupèrer est la longueur de la chaine divisée
		// par 20 car chaque puck est codé sur 20 caractères
		int nb = s.length() / 20;
		Puck[] reception = new Puck[nb];

		// pour tous les pucks à recuperer
		for (int i = 0; i < reception.length; i++) {
			// on recupère la ieme chaine de 20 caractères
			reception[i] = puckFromString(s.substring(20 * i, 20 * i + 20));
		}

		// retour bidon
		return reception;
	}

}
